"""
This script aims to convert nuplan scenarios to ScenarioDescription, so that we can load any nuplan scenarios into
MetaDrive.
"""
import copy
import os
import pickle
import shutil

import tqdm

from metadrive.engine.asset_loader import AssetLoader
from metadrive.scenario.scenario_description import ScenarioDescription
from metadrive.utils.nuplan.utils import get_nuplan_scenarios, convert_one_scenario
from metadrive.utils.utils import dict_recursive_remove_array

import logging

logger = logging.getLogger(__name__)

logger.warning(
    "This converters will de deprecated. "
    "Use tools in ScenarioNet instead: https://github.com/metadriverse/ScenarioNet."
)


def convert_scenarios(output_path, dataset_params, worker_index=None, force_overwrite=False):
    save_path = copy.deepcopy(output_path)
    output_path = output_path + "_tmp"
    # meta recorder and data summary
    if os.path.exists(output_path):
        shutil.rmtree(output_path)
    os.makedirs(output_path, exist_ok=False)

    # make real save dir
    delay_remove = None
    if os.path.exists(save_path):
        if force_overwrite:
            delay_remove = save_path
        else:
            raise ValueError("Directory already exists! Abort")

    metadata_recorder = {}
    total_scenarios = 0
    desc = ""
    summary_file = ScenarioDescription.DATASET.SUMMARY_FILE
    if worker_index is not None:
        desc += "Worker {} ".format(worker_index)
        summary_file = "dataset_summary_worker{}.pkl".format(worker_index)

    # Init.
    scenarios = get_nuplan_scenarios(dataset_params)
    for scenario in tqdm.tqdm(scenarios):
        sd_scenario = convert_one_scenario(scenario)
        sd_scenario = sd_scenario.to_dict()
        ScenarioDescription.sanity_check(sd_scenario, check_self_type=True)
        export_file_name = ScenarioDescription.get_export_file_name("nuplan", "v1.1", scenario.scenario_name)
        p = os.path.join(output_path, export_file_name)
        with open(p, "wb") as f:
            pickle.dump(sd_scenario, f)
        metadata_recorder[export_file_name] = copy.deepcopy(sd_scenario[ScenarioDescription.METADATA])
    # rename and save
    if delay_remove is not None:
        shutil.rmtree(delay_remove)
    os.rename(output_path, save_path)
    summary_file = os.path.join(save_path, summary_file)
    with open(summary_file, "wb") as file:
        pickle.dump(dict_recursive_remove_array(metadata_recorder), file)
    print("Summary is saved at: {}".format(summary_file))
    if delay_remove is not None:
        assert delay_remove == save_path, delay_remove + " vs. " + save_path


if __name__ == "__main__":
    # 14 types
    all_scenario_types = "[behind_pedestrian_on_pickup_dropoff,  \
                            near_multiple_vehicles, \
                            high_magnitude_jerk, \
                            crossed_by_vehicle, \
                            following_lane_with_lead, \
                            changing_lane_to_left, \
                            accelerating_at_traffic_light_without_lead, \
                            stopping_at_stop_sign_with_lead, \
                            traversing_narrow_lane, \
                            waiting_for_pedestrian_to_cross, \
                            starting_left_turn, \
                            starting_high_speed_turn, \
                            starting_unprotected_cross_turn, \
                            starting_protected_noncross_turn, \
                            on_pickup_dropoff]"

    dataset_params = [
        # builder setting
        "scenario_builder=nuplan_mini",
        "scenario_builder.scenario_mapping.subsample_ratio_override=0.5",  # 10 hz

        # filter
        "scenario_filter=all_scenarios",  # simulate only one log
        "scenario_filter.remove_invalid_goals=true",
        "scenario_filter.shuffle=true",
        "scenario_filter.log_names=['2021.07.16.20.45.29_veh-35_01095_01486']",
        # "scenario_filter.scenario_types={}".format(all_scenario_types),
        # "scenario_filter.scenario_tokens=[]",
        # "scenario_filter.map_names=[]",
        # "scenario_filter.num_scenarios_per_type=1",
        # "scenario_filter.limit_total_scenarios=1000",
        # "scenario_filter.expand_scenarios=true",
        # "scenario_filter.limit_scenarios_per_type=10",  # use 10 scenarios per scenario type
        "scenario_filter.timestamp_threshold_s=20",  # minial scenario duration (s)
    ]
    output_path = AssetLoader.file_path("nuplan", return_raw_style=False)
    worker_index = None
    force_overwrite = True
    convert_scenarios(output_path, dataset_params, worker_index=worker_index, force_overwrite=force_overwrite)



import numpy as np
import cv2

from metadrive import MetaDriveEnv


def _test_main_camera_as_obs(render):
    try:
        env = MetaDriveEnv(
            dict(
                num_scenarios=1000,
                start_seed=1010,
                traffic_density=0.05,
                image_observation=True,
                use_render=False,
                vehicle_config=dict(image_source="main_camera"),
                show_interface=False,
                show_logo=False,
                show_fps=False,
            )
        )
        obs, _ = env.reset()
        action = [0.0, 1.]
        for _ in range(3):
            env.reset()
            for s in range(20):
                o, r, tm, tc, i = env.step(action)
                # engine = env.engine
                # if engine.episode_step <= 1:
                #     engine.graphicsEngine.renderFrame()
                # origin_img = engine.win.getDisplayRegion(0).getScreenshot()
                # v = memoryview(origin_img.getRamImage).tolist()
                # img = np.array(v)
                # img = img.reshape((origin_img.getYSize(), origin_img.getXSize(), 4))
                # img = img[::-1]
                # img = img[..., :-1]
                # img = img/255
                assert np.sum(o["image"][..., -1]) > 10
                assert not env.engine.interface.need_interface
                if render:
                    cv2.imshow("window", o["image"][..., -1])
                    cv2.waitKey(1)
    finally:
        env.close()


def _test_rgb_camera_as_obs():
    try:
        env = MetaDriveEnv(
            dict(
                num_scenarios=1000,
                start_seed=1010,
                traffic_density=0.05,
                image_observation=True,
                use_render=False,
                vehicle_config=dict(image_source="rgb_camera"),
                show_interface=False,
                show_logo=False,
                show_fps=False,
            )
        )
        obs, _ = env.reset()
        action = [0.0, 1.]
        for _ in range(3):
            env.reset()
            for s in range(20):
                o, r, tm, tc, i = env.step(action)
                # engine = env.engine
                # if engine.episode_step <= 1:
                #     engine.graphicsEngine.renderFrame()
                # origin_img = engine.win.getDisplayRegion(0).getScreenshot()
                # v = memoryview(origin_img.getRamImage).tolist()
                # img = np.array(v)
                # img = img.reshape((origin_img.getYSize(), origin_img.getXSize(), 4))
                # img = img[::-1]
                # img = img[..., :-1]
                # img = img/255
                assert np.sum(o["image"][..., -1]) > 10
                # cv2.waitKey(1)
    finally:
        env.close()


if __name__ == "__main__":
    _test_main_camera_as_obs(True)



"""

Renders the sphere using the render pipeline

"""

import _tmp_material as material

import sys
from panda3d.core import load_prc_file_data, Vec4
from direct.showbase.ShowBase import ShowBase


class Application(ShowBase):
    def __init__(self):
        sys.path.insert(0, "../../")
        load_prc_file_data("", "win-size 512 512")
        load_prc_file_data("", "textures-power-2 none")
        load_prc_file_data("", "print-pipe-types #f")
        load_prc_file_data("", "notify-level-glgsg error")
        # load_prc_file_data("", "win-size 1024 1024")

        from rpcore import RenderPipeline, PointLight

        self.render_pipeline = RenderPipeline()
        self.render_pipeline.mount_mgr.config_dir = "config/"
        self.render_pipeline.create(self)

        sphere = self.loader.loadModel("res/sphere.bam")
        sphere.reparent_to(self.render)

        self.disableMouse()
        self.camLens.setFov(40)
        self.camLens.setNearFar(0.03, 2000.0)
        self.camera.set_pos(0, -3.5, 0)
        self.camera.look_at(0, -2.5, 0)

        self.render2d.hide()
        self.aspect2d.hide()

        light = PointLight()
        light.pos = 10, -10, 10
        light.radius = 1e20
        light.color = (1, 1, 1)
        light.inner_radius = 4.0
        light.energy = 3
        self.render_pipeline.add_light(light)

        light = PointLight()
        light.pos = -10, -10, 10
        light.radius = 1e20
        light.color = (1, 1, 1)
        light.inner_radius = 4.0
        light.energy = 3
        self.render_pipeline.add_light(light)

        for mat in sphere.find_all_materials():
            mat.roughness = material.roughness
            mat.base_color = Vec4(*(list(material.basecolor) + [1]))
            mat.refractive_index = material.ior

            mat.metallic = 1.0 if material.mat_type == "metallic" else 0.0

            if material.mat_type == "clearcoat":
                mat.emission = (2, 0, 0, 0)
                mat.metallic = 1.0
                mat.refractive_index = 1.51

            if material.mat_type == "foliage":
                mat.emission = (5, 0, 0, 0)
                mat.metallic = 0.0
                mat.refractive_index = 1.51

        for i in range(10):
            self.taskMgr.step()

        self.win.save_screenshot("scene-rp.png")

        self.accept("r", self.reload)

    def reload(self):
        print("Reloading")
        self.render_pipeline.reload_shaders()

        for i in range(4):
            self.taskMgr.step()

        self.win.save_screenshot("scene-rp.png")


if len(sys.argv) <= 1:
    Application().run()
else:
    Application()



from metadrive.constants import MetaDriveType
from metadrive.utils.utils import get_object_from_node


def collision_callback(contact):
    """
    All collision callback should be here, and a notify() method can turn it on
    It may lower the performance if overdone
    """

    # now it only process BaseVehicle collision
    node0 = contact.getNode0()
    node1 = contact.getNode1()

    nodes = [node0, node1]
    another_nodes = [node1, node0]
    for i in range(2):
        if nodes[i].hasPythonTag(MetaDriveType.VEHICLE):
            obj_1 = get_object_from_node(nodes[i])
            obj_2 = get_object_from_node(another_nodes[i])
            another_node_name = another_nodes[i].getName()
            # crash vehicles
            if another_node_name == MetaDriveType.VEHICLE:
                obj_1.crash_vehicle = True
            # crash objects
            elif MetaDriveType.is_traffic_object(another_node_name):
                if not obj_2.crashed:
                    obj_1.crash_object = True
                    if obj_2.COST_ONCE:
                        obj_2.crashed = True
            # collision_human
            elif another_node_name in [MetaDriveType.CYCLIST, MetaDriveType.PEDESTRIAN]:
                obj_1.crash_human = True
            # crash invisible wall or building
            elif another_node_name in [MetaDriveType.INVISIBLE_WALL, MetaDriveType.BUILDING]:
                obj_1.crash_building = True
            # logging.debug("{} crash with {}".format(nodes[i].getName(), another_nodes[i].getName()))



import osmnx as ox
from pyrosm import OSM

# get real map in PBF fromat from https://extract.bbbike.org/
# install pyrosm and all dependencies via pip install pyrosm osmnx networkx

# Initialize the reader
osm = OSM("cpii.osm.pbf")

# Get all walkable roads and the nodes
nodes, edges = osm.get_network(nodes=True, network_type="driving")
# Create NetworkX graph
G = osm.to_graph(nodes, edges, graph_type="networkx")

ccc = edges.head(40)
# print(ccc.geometry)
ox.plot_graph(G)



class PIDController:
    def __init__(self, k_p: float, k_i: float, k_d: float):
        self.k_p = k_p
        self.k_i = k_i
        self.k_d = k_d
        self.p_error = 0
        self.i_error = 0
        self.d_error = 0

    def _update_error(self, current_error: float):
        self.i_error += current_error
        self.d_error = current_error - self.p_error
        self.p_error = current_error

    def get_result(self, current_error: float, make_up_coefficient=1.0):
        self._update_error(current_error)
        return (-self.k_p * self.p_error - self.k_i * self.i_error - self.k_d * self.d_error) * make_up_coefficient

    def reset(self):
        self.p_error = 0
        self.i_error = 0
        self.d_error = 0


class Target:
    def __init__(self, target_lateral, target_speed_km_h):
        """
        target speed: Km/h
        """
        self.lateral = target_lateral
        self.speed_km_h = target_speed_km_h

    def go_right(self):
        self.lateral += 0.25 if self.lateral < 0.625 else 0

    def go_left(self):
        self.lateral -= 0.25 if self.lateral > 0.125 else 0

    def faster(self):
        self.speed_km_h += 10

    def slower(self):
        self.speed_km_h -= 10



from panda3d.core import Vec3

from metadrive.component.sensors.base_camera import BaseCamera
from metadrive.constants import CamMask
from metadrive.engine.engine_utils import get_global_config, engine_initialized


class MiniMap(BaseCamera):
    CAM_MASK = CamMask.MiniMap
    display_region_size = [0., 1 / 3, 0.8, 1.0]

    def __init__(self, engine, width, height, z_pos, *, cuda=False):
        self.BUFFER_W, self.BUFFER_H, height = width, height, z_pos
        super(MiniMap, self).__init__(engine=engine, need_cuda=cuda)

        cam = self.get_cam()
        lens = self.get_lens()

        cam.setZ(height)
        cam.lookAt(Vec3(20, 0, 0))
        lens.setAspectRatio(2.0)



"""

RenderPipeline

Copyright (c) 2014-2016 tobspr <tobias.springer1@gmail.com>

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.

"""

from __future__ import division
from panda3d.core import Vec4

from metadrive.render_pipeline.rpcore.render_stage import RenderStage
from metadrive.render_pipeline.rpcore.globals import Globals
from metadrive.render_pipeline.rpcore.image import Image


class AutoExposureStage(RenderStage):

    required_pipes = ["ShadedScene"]
    required_inputs = []

    @property
    def produced_pipes(self):
        return {"ShadedScene": self.target_apply.color_tex, "Exposure": self.tex_exposure}

    def create(self):

        # Create the target which converts the scene color to a luminance
        self.target_lum = self.create_target("GetLuminance")
        self.target_lum.size = -4
        self.target_lum.add_color_attachment(bits=(16, 0, 0, 0))
        self.target_lum.prepare_buffer()

        self.mip_targets = []

        # Create the storage for the exposure, this stores the current and last
        # frames exposure
        # XXX: We have to use F_r16 instead of F_r32 because of a weird nvidia
        # driver bug! However, 16 bits should be enough for sure.
        self.tex_exposure = Image.create_buffer("ExposureStorage", 1, "R16")
        self.tex_exposure.set_clear_color(Vec4(0.5))
        self.tex_exposure.clear_image()

        # Create the target which extracts the exposure from the average brightness
        self.target_analyze = self.create_target("AnalyzeBrightness")
        self.target_analyze.size = 1, 1
        self.target_analyze.prepare_buffer()

        self.target_analyze.set_shader_input("ExposureStorage", self.tex_exposure)

        # Create the target which applies the generated exposure to the scene
        self.target_apply = self.create_target("ApplyExposure")
        self.target_apply.add_color_attachment(bits=16)
        self.target_apply.prepare_buffer()
        self.target_apply.set_shader_input("Exposure", self.tex_exposure)

    def set_dimensions(self):
        for old_target in self.mip_targets:
            self.remove_target(old_target)

        wsize_x = (Globals.resolution.x + 3) // 4
        wsize_y = (Globals.resolution.y + 3) // 4

        # Create the targets which downscale the luminance mipmaps
        self.mip_targets = []
        last_tex = self.target_lum.color_tex
        while wsize_x >= 4 or wsize_y >= 4:
            wsize_x = (wsize_x + 3) // 4
            wsize_y = (wsize_y + 3) // 4

            mip_target = self.create_target("DScaleLum:S" + str(wsize_x))
            mip_target.add_color_attachment(bits=(16, 0, 0, 0))
            mip_target.size = wsize_x, wsize_y
            mip_target.sort = self.target_lum.sort + len(self.mip_targets)
            mip_target.prepare_buffer()
            mip_target.set_shader_input("SourceTex", last_tex)
            self.mip_targets.append(mip_target)
            last_tex = mip_target.color_tex

        self.target_analyze.set_shader_input("DownscaledTex", self.mip_targets[-1].color_tex)

        # Shaders might not have been loaded at this point
        if hasattr(self, "mip_shader"):
            for target in self.mip_targets:
                target.shader = self.mip_shader

    def reload_shaders(self):
        self.target_lum.shader = self.load_plugin_shader("generate_luminance.frag.glsl")
        self.target_analyze.shader = self.load_plugin_shader("analyze_brightness.frag.glsl")
        self.target_apply.shader = self.load_plugin_shader("apply_exposure.frag.glsl")

        # Keep shader as reference, required when resizing
        self.mip_shader = self.load_plugin_shader("downscale_luminance.frag.glsl")
        for target in self.mip_targets:
            target.shader = self.mip_shader



"""
This filed is mostly copied from gym==0.17.2
We use the gym.spaces as helpers, but it may cause problem if user using some old version of gym.
"""

import logging
import typing as tp
from collections import namedtuple, OrderedDict

import numpy as np

from metadrive.utils import get_np_random

BoxSpace = namedtuple("BoxSpace", "max min")
DiscreteSpace = namedtuple("DiscreteSpace", "max min")
ConstantSpace = namedtuple("ConstantSpace", "value")


class Space:
    """
    Copied from gym: gym/spaces/space.py

    Defines the observation and action spaces, so you can write generic
    code that applies to any Env. For example, you can choose a random
    action.
    """
    def __init__(self, shape=None, dtype=None):
        import numpy as np  # takes about 300-400ms to import, so we load lazily
        self.shape = None if shape is None else tuple(shape)
        self.dtype = None if dtype is None else np.dtype(dtype)
        self.np_random = None
        self.seed()

    def sample(self):
        """Randomly sample an element of this space. Can be
        uniform or non-uniform sampling based on boundedness of space."""
        raise NotImplementedError

    def seed(self, seed=None):
        """Seed the PRNG of this space. """
        self.np_random, seed = get_np_random(seed, return_seed=True)
        return [seed]

    def contains(self, x):
        """
        Return boolean specifying if x is a valid
        member of this space
        """
        raise NotImplementedError

    def __contains__(self, x):
        return self.contains(x)

    def to_jsonable(self, sample_n):
        """Convert a batch of samples from this space to a JSONable data type."""
        # By default, assume identity is JSONable
        return sample_n

    def from_jsonable(self, sample_n):
        """Convert a JSONable data type to a batch of samples from this space."""
        # By default, assume identity is JSONable
        return sample_n

    def __del__(self):
        del self.np_random


class Dict(Space):
    """
    Copied from gym: gym/spaces/dcit.py

    A dictionary of simpler spaces.

    Example usage:
    self.observation_space = spaces.Dict({"position": spaces.Discrete(2), "velocity": spaces.Discrete(3)})

    Example usage [nested]:
    self.nested_observation_space = spaces.Dict({
        'sensors':  spaces.Dict({
            'position': spaces.Box(low=-100, high=100, shape=(3,)),
            'velocity': spaces.Box(low=-1, high=1, shape=(3,)),
            'front_cam': spaces.Tuple((
                spaces.Box(low=0, high=1, shape=(10, 10, 3)),
                spaces.Box(low=0, high=1, shape=(10, 10, 3))
            )),
            'rear_cam': spaces.Box(low=0, high=1, shape=(10, 10, 3)),
        }),
        'ext_controller': spaces.MultiDiscrete((5, 2, 2)),
        'inner_state':spaces.Dict({
            'charge': spaces.Discrete(100),
            'system_checks': spaces.MultiBinary(10),
            'job_status': spaces.Dict({
                'task': spaces.Discrete(5),
                'progress': spaces.Box(low=0, high=100, shape=()),
            })
        })
    })
    """
    def __init__(self, spaces=None, **spaces_kwargs):
        assert (spaces is None) or (not spaces_kwargs), 'Use either Dict(spaces=dict(...)) or Dict(foo=x, bar=z)'
        if spaces is None:
            spaces = spaces_kwargs
        if isinstance(spaces, dict) and not isinstance(spaces, OrderedDict):
            spaces = OrderedDict(sorted(list(spaces.items())))
        if isinstance(spaces, list):
            spaces = OrderedDict(spaces)
        self.spaces = spaces
        for space in spaces.values():
            assert isinstance(space, Space), 'Values of the dict should be instances of gym.Space'
        super(Dict, self).__init__(None, None)  # None for shape and dtype, since it'll require special handling

    def seed(self, seed=None):
        for space in self.spaces.values():
            space.seed(seed)

    def sample(self):
        return OrderedDict([(k, space.sample()) for k, space in self.spaces.items()])

    def contains(self, x):
        if not isinstance(x, dict) or len(x) != len(self.spaces):
            return False
        for k, space in self.spaces.items():
            if k not in x:
                return False
            if not space.contains(x[k]):
                return False
        return True

    def __getitem__(self, key):
        return self.spaces[key]

    def __repr__(self):
        return "Dict(" + ", ".join([str(k) + ":" + str(s) for k, s in self.spaces.items()]) + ")"

    def to_jsonable(self, sample_n):
        # serialize as dict-repr of vectors
        return {key: space.to_jsonable([sample[key] for sample in sample_n]) \
                for key, space in self.spaces.items()}

    def from_jsonable(self, sample_n):
        dict_of_list = {}
        for key, space in self.spaces.items():
            dict_of_list[key] = space.from_jsonable(sample_n[key])
        ret = []
        for i, _ in enumerate(dict_of_list[key]):
            entry = {}
            for key, value in dict_of_list.items():
                entry[key] = value[i]
            ret.append(entry)
        return ret

    def __eq__(self, other):
        return isinstance(other, Dict) and self.spaces == other.spaces


class ParameterSpace(Dict):
    """
    length = PGSpace(name="length",max=50.0,min=10.0)
    Usage:
    PGSpace({"lane_length":length})
    """
    def __init__(self, our_config: tp.Dict[str, tp.Union[BoxSpace, DiscreteSpace, ConstantSpace]]):
        super(ParameterSpace, self).__init__(ParameterSpace.wrap2gym_space(our_config))
        self.parameters = set(our_config.keys())

    @staticmethod
    def wrap2gym_space(our_config):
        ret = dict()
        for key, value in our_config.items():
            if isinstance(value, BoxSpace):
                ret[key] = Box(low=value.min, high=value.max, shape=(1, ))
            elif isinstance(value, DiscreteSpace):
                ret[key] = Box(low=value.min, high=value.max, shape=(1, ), dtype=np.int64)
            elif isinstance(value, ConstantSpace):
                ret[key] = Box(low=value.value, high=value.value, shape=(1, ))
            else:
                raise ValueError("{} can not be wrapped in gym space".format(key))
        return ret


class Parameter:
    """
    Block parameters and vehicle parameters
    """
    # block
    length = "length"
    radius = "radius"
    angle = "angle"
    goal = "goal"
    dir = "dir"
    radius_inner = "inner_radius"  # only for roundabout use
    radius_exit = "exit_radius"
    exit_length = "exit_length"  # The length of the exit parts straight lane, for roundabout use only.
    t_intersection_type = "t_type"
    lane_num = "lane_num"
    change_lane_num = "change_lane_num"
    decrease_increase = "decrease_increase"
    one_side_vehicle_num = "one_side_vehicle_number"

    # vehicle
    # vehicle_length = "v_len"
    # vehicle_width = "v_width"
    vehicle_height = "v_height"
    front_tire_longitude = "f_tire_long"
    rear_tire_longitude = "r_tire_long"
    tire_lateral = "tire_lateral"
    tire_axis_height = "tire_axis_height"
    tire_radius = "tire_radius"
    mass = "mass"  # kg
    heading = "heading"
    # steering_max = "steering_max"
    # engine_force_max = "e_f_max"
    # brake_force_max = "b_f_max"
    # speed_max = "s_max"

    # vehicle visualization
    vehicle_vis_z = "vis_z"
    vehicle_vis_y = "vis_y"
    vehicle_vis_h = "vis_h"
    vehicle_vis_scale = "vis_scale"


class VehicleParameterSpace:
    STATIC_BASE_VEHICLE = dict(
        wheel_friction=ConstantSpace(0.9),
        max_engine_force=ConstantSpace(800),
        max_brake_force=ConstantSpace(150),
        max_steering=ConstantSpace(40),
        max_speed_km_h=ConstantSpace(80),
    )
    STATIC_DEFAULT_VEHICLE = STATIC_BASE_VEHICLE

    BASE_VEHICLE = dict(
        wheel_friction=ConstantSpace(0.9),
        max_engine_force=BoxSpace(750, 850),
        max_brake_force=BoxSpace(80, 180),
        max_steering=ConstantSpace(40),
        max_speed_km_h=ConstantSpace(80),
    )
    DEFAULT_VEHICLE = BASE_VEHICLE

    S_VEHICLE = dict(
        wheel_friction=ConstantSpace(0.9),
        max_engine_force=BoxSpace(350, 550),
        max_brake_force=BoxSpace(35, 80),
        max_steering=ConstantSpace(50),
        max_speed_km_h=ConstantSpace(80),
    )
    M_VEHICLE = dict(
        wheel_friction=ConstantSpace(0.75),
        max_engine_force=BoxSpace(650, 850),
        max_brake_force=BoxSpace(60, 150),
        max_steering=ConstantSpace(45),
        max_speed_km_h=ConstantSpace(80),
    )
    L_VEHICLE = dict(
        wheel_friction=ConstantSpace(0.8),
        max_engine_force=BoxSpace(450, 650),
        max_brake_force=BoxSpace(60, 120),
        max_steering=ConstantSpace(40),
        max_speed_km_h=ConstantSpace(80),
    )
    XL_VEHICLE = dict(
        wheel_friction=ConstantSpace(0.7),
        max_engine_force=BoxSpace(500, 700),
        max_brake_force=BoxSpace(50, 100),
        max_steering=ConstantSpace(35),
        max_speed_km_h=ConstantSpace(80),
    )


class BlockParameterSpace:
    """
    Make sure the range of curve parameters covers the parameter space of other blocks,
    otherwise, an error may happen in navigation info normalization
    """
    STRAIGHT = {Parameter.length: BoxSpace(min=40.0, max=80.0)}
    BIDIRECTION = {Parameter.length: BoxSpace(min=40.0, max=80.0)}

    CURVE = {
        Parameter.length: BoxSpace(min=40.0, max=80.0),
        Parameter.radius: BoxSpace(min=25.0, max=60.0),
        Parameter.angle: BoxSpace(min=45, max=135),
        Parameter.dir: DiscreteSpace(min=0, max=1)
    }
    INTERSECTION = {
        Parameter.radius: ConstantSpace(10),
        Parameter.change_lane_num: DiscreteSpace(min=0, max=1),  # 0, 1
        Parameter.decrease_increase: DiscreteSpace(min=0, max=1)  # 0, decrease, 1 increase
    }
    ROUNDABOUT = {
        # The radius of the
        Parameter.radius_exit: BoxSpace(min=5, max=15),
        Parameter.radius_inner: BoxSpace(min=15, max=45),
        Parameter.angle: ConstantSpace(60)
    }
    T_INTERSECTION = {
        Parameter.radius: ConstantSpace(10),
        Parameter.t_intersection_type: DiscreteSpace(min=0, max=2),  # 3 different t type for previous socket
        Parameter.change_lane_num: DiscreteSpace(min=0, max=1),  # 0,1
        Parameter.decrease_increase: DiscreteSpace(min=0, max=1)  # 0, decrease, 1 increase
    }
    RAMP_PARAMETER = {
        Parameter.length: BoxSpace(min=20, max=40)  # accelerate/decelerate part length
    }
    FORK_PARAMETER = {
        Parameter.length: BoxSpace(min=20, max=40),  # accelerate/decelerate part length
        Parameter.lane_num: DiscreteSpace(min=0, max=1)
    }
    BOTTLENECK_PARAMETER = {
        Parameter.length: BoxSpace(min=20, max=50),  # the length of straigh part
        Parameter.lane_num: DiscreteSpace(min=1, max=2),  # the lane num increased or decreased now 1-2
        "bottle_len": ConstantSpace(20),
        "solid_center_line": ConstantSpace(0)  # bool, turn on yellow line or not
    }
    TOLLGATE_PARAMETER = {
        Parameter.length: ConstantSpace(20),  # the length of straigh part
    }
    PARKING_LOT_PARAMETER = {
        Parameter.one_side_vehicle_num: DiscreteSpace(min=2, max=10),
        Parameter.radius: ConstantSpace(value=4),
        Parameter.length: ConstantSpace(value=8)
    }


class Discrete(Space):
    r"""
    Copied from gym: gym/spaces/discrete.py

    A discrete space in :math:`\{ 0, 1, \\dots, n-1 \}`.

    Example::

        >>> Discrete(2)

    """
    def __init__(self, n):
        assert n >= 0
        self.n = n
        super(Discrete, self).__init__((), np.int64)

    def sample(self):
        return self.np_random.randint(self.n)

    def contains(self, x):
        if isinstance(x, int):
            as_int = x
        elif isinstance(x, (np.generic, np.ndarray)) and (x.dtype.char in np.typecodes['AllInteger'] and x.shape == ()):
            as_int = int(x)
        else:
            return False
        return as_int >= 0 and as_int < self.n

    def __repr__(self):
        return "Discrete(%d)" % self.n

    def __eq__(self, other):
        return isinstance(other, Discrete) and self.n == other.n


class Box(Space):
    """
    Copied from gym: gym/spaces/box.py

    A (possibly unbounded) box in R^n. Specifically, a Box represents the
    Cartesian product of n closed intervals. Each interval has the form of one
    of [a, b], (-oo, b], [a, oo), or (-oo, oo).

    There are two common use cases:

    * Identical bound for each dimension::
        >>> Box(low=-1.0, high=2.0, shape=(3, 4), dtype=np.float32)
        Box(3, 4)

    * Independent bound for each dimension::
        >>> Box(low=np.array([-1.0, -2.0]), high=np.array([2.0, 4.0]), dtype=np.float32)
        Box(2,)

    """
    def __init__(self, low, high, shape=None, dtype=np.float32):
        assert dtype is not None, 'dtype must be explicitly provided. '
        self.dtype = np.dtype(dtype)

        # determine shape if it isn't provided directly
        if shape is not None:
            shape = tuple(shape)
            assert np.isscalar(low) or low.shape == shape, "low.shape doesn't match provided shape"
            assert np.isscalar(high) or high.shape == shape, "high.shape doesn't match provided shape"
        elif not np.isscalar(low):
            shape = low.shape
            assert np.isscalar(high) or high.shape == shape, "high.shape doesn't match low.shape"
        elif not np.isscalar(high):
            shape = high.shape
            assert np.isscalar(low) or low.shape == shape, "low.shape doesn't match high.shape"
        else:
            raise ValueError("shape must be provided or inferred from the shapes of low or high")

        if np.isscalar(low):
            low = np.full(shape, low, dtype=dtype)

        if np.isscalar(high):
            high = np.full(shape, high, dtype=dtype)

        self.shape = shape
        self.low = low
        self.high = high

        def _get_precision(dtype):
            if np.issubdtype(dtype, np.floating):
                return np.finfo(dtype).precision
            else:
                return np.inf

        low_precision = _get_precision(self.low.dtype)
        high_precision = _get_precision(self.high.dtype)
        dtype_precision = _get_precision(self.dtype)
        if min(low_precision, high_precision) > dtype_precision:
            logging.warning("Box bound precision lowered by casting to {}".format(self.dtype))
        self.low = self.low.astype(self.dtype)
        self.high = self.high.astype(self.dtype)

        # Boolean arrays which indicate the interval type for each coordinate
        self.bounded_below = -np.inf < self.low
        self.bounded_above = np.inf > self.high

        super(Box, self).__init__(self.shape, self.dtype)

    def is_bounded(self, manner="both"):
        below = np.all(self.bounded_below)
        above = np.all(self.bounded_above)
        if manner == "both":
            return below and above
        elif manner == "below":
            return below
        elif manner == "above":
            return above
        else:
            raise ValueError("manner is not in {'below', 'above', 'both'}")

    def sample(self):
        """
        Generates a single random sample inside of the Box.

        In creating a sample of the box, each coordinate is sampled according to
        the form of the interval:

        * [a, b] : uniform distribution
        * [a, oo) : shifted exponential distribution
        * (-oo, b] : shifted negative exponential distribution
        * (-oo, oo) : normal distribution
        """
        high = self.high if self.dtype.kind == 'f' \
            else self.high.astype('int64') + 1
        sample = np.empty(self.shape)

        # Masking arrays which classify the coordinates according to interval
        # type
        unbounded = ~self.bounded_below & ~self.bounded_above
        upp_bounded = ~self.bounded_below & self.bounded_above
        low_bounded = self.bounded_below & ~self.bounded_above
        bounded = self.bounded_below & self.bounded_above

        # Vectorized sampling by interval type
        sample[unbounded] = self.np_random.normal(size=unbounded[unbounded].shape)

        sample[low_bounded] = self.np_random.exponential(size=low_bounded[low_bounded].shape) + self.low[low_bounded]

        sample[upp_bounded] = -self.np_random.exponential(size=upp_bounded[upp_bounded].shape) + self.high[upp_bounded]

        sample[bounded] = self.np_random.uniform(low=self.low[bounded], high=high[bounded], size=bounded[bounded].shape)
        if self.dtype.kind == 'i':
            sample = np.floor(sample)

        return sample.astype(self.dtype)

    def contains(self, x):
        if isinstance(x, list):
            x = np.array(x)  # Promote list to array for contains check
        return x.shape == self.shape and np.all(x >= self.low) and np.all(x <= self.high)

    def to_jsonable(self, sample_n):
        return np.array(sample_n).tolist()

    def from_jsonable(self, sample_n):
        return [np.asarray(sample) for sample in sample_n]

    def __repr__(self):
        return "Box" + str(self.shape)

    def __eq__(self, other):
        return isinstance(other, Box) and \
               (self.shape == other.shape) and \
               np.allclose(self.low, other.low) and \
               np.allclose(self.high, other.high)


if __name__ == "__main__":
    """
    Test
    """
    config = {
        "length": BoxSpace(min=10.0, max=80.0),
        "angle": BoxSpace(min=50.0, max=360.0),
        "goal": DiscreteSpace(min=0, max=2)
    }
    config = ParameterSpace(config)
    print(config.sample())
    config.seed(1)
    print(config.sample())
    print(config.sample())
    config.seed(1)
    print(*config.sample()["length"])



import copy
import logging
from metadrive.utils.math import norm
import numpy as np

from metadrive.component.static_object.traffic_object import TrafficCone, TrafficBarrier
from metadrive.component.traffic_participants.cyclist import Cyclist
from metadrive.component.traffic_participants.pedestrian import Pedestrian
from metadrive.component.vehicle.vehicle_type import get_vehicle_type, reset_vehicle_type_count
from metadrive.constants import DEFAULT_AGENT
from metadrive.manager.base_manager import BaseManager
from metadrive.policy.idm_policy import ScenarioIDMPolicy
from metadrive.policy.replay_policy import ReplayEgoCarPolicy
from metadrive.policy.replay_policy import ReplayTrafficParticipantPolicy
from metadrive.scenario.parse_object_state import parse_object_state, get_idm_route, get_max_valid_indicis
from metadrive.scenario.scenario_description import ScenarioDescription as SD
from metadrive.type import MetaDriveType
from metadrive.utils.math import wrap_to_pi

logger = logging.getLogger(__name__)


class ScenarioTrafficManager(BaseManager):
    STATIC_THRESHOLD = 3  # m, static if moving distance < 5
    IDM_ACT_BATCH_SIZE = 5

    # project cars to ego vehicle coordinates, only vehicles behind ego car and in a certain region can get IDM policy
    IDM_CREATE_SIDE_CONSTRAINT = 15  # m
    IDM_CREATE_FORWARD_CONSTRAINT = -1  # m
    IDM_CREATE_MIN_LENGTH = 5  # m

    # project cars to ego vehicle coordinates, only vehicles outside the region can be created
    GENERATION_SIDE_CONSTRAINT = 2  # m
    GENERATION_FORWARD_CONSTRAINT = 8  # m

    # filter noise static object: barrier and cone
    MIN_VALID_FRAME_LEN = 20  # frames

    def __init__(self):
        super(ScenarioTrafficManager, self).__init__()
        self._scenario_id_to_obj_id = None
        self._obj_id_to_scenario_id = None

        # for filtering some static cars
        self._static_car_id = set()
        self._moving_car_id = set()

        # for filtering noisy static object
        self._noise_object_id = set()
        self._non_noise_object_id = set()

        # an async trick for accelerating IDM policy
        self.idm_policy_count = 0
        self._obj_to_clean_this_frame = []

        # some flags
        self.even_sample_v = self.engine.global_config["even_sample_vehicle_class"]
        self.need_default_vehicle = self.engine.global_config["default_vehicle_in_traffic"]
        self.is_ego_vehicle_replay = self.engine.global_config["agent_policy"] == ReplayEgoCarPolicy
        self._filter_overlapping_car = self.engine.global_config["filter_overlapping_car"]

        # config
        self._traffic_v_config = self.get_traffic_v_config()

    def before_step(self, *args, **kwargs):
        self._obj_to_clean_this_frame = []
        for v in self.spawned_objects.values():
            if self.engine.has_policy(v.id, ScenarioIDMPolicy):
                p = self.engine.get_policy(v.name)
                if p.arrive_destination:
                    self._obj_to_clean_this_frame.append(self._obj_id_to_scenario_id[v.id])
                else:
                    do_speed_control = self.episode_step % self.IDM_ACT_BATCH_SIZE == p.policy_index
                    v.before_step(p.act(do_speed_control))

    def before_reset(self):
        super(ScenarioTrafficManager, self).before_reset()
        self._obj_to_clean_this_frame = []
        reset_vehicle_type_count(self.np_random)

    def after_reset(self):
        self._scenario_id_to_obj_id = {}
        self._obj_id_to_scenario_id = {}
        self._static_car_id = set()
        self._moving_car_id = set()
        self._noise_object_id = set()
        self._non_noise_object_id = set()
        self.idm_policy_count = 0
        for scenario_id, track in self.current_traffic_data.items():
            if scenario_id == self.sdc_scenario_id:
                continue
            if track["type"] == MetaDriveType.VEHICLE:
                self.spawn_vehicle(scenario_id, track)
            elif track["type"] == MetaDriveType.CYCLIST:
                self.spawn_cyclist(scenario_id, track)
            elif track["type"] == MetaDriveType.PEDESTRIAN:
                self.spawn_pedestrian(scenario_id, track)
            elif track["type"] in [MetaDriveType.TRAFFIC_CONE, MetaDriveType.TRAFFIC_BARRIER]:
                cls = TrafficBarrier if track["type"] == MetaDriveType.TRAFFIC_BARRIER else TrafficCone
                self.spawn_static_object(cls, scenario_id, track)
            else:
                logger.warning("Do not support {}".format(track["type"]))

    def after_step(self, *args, **kwargs):
        if self.episode_step < self.current_scenario_length:
            replay_done = False
            for scenario_id, track in self.current_traffic_data.items():
                if scenario_id == self.sdc_scenario_id:
                    continue
                if scenario_id not in self._scenario_id_to_obj_id:
                    if track["type"] == MetaDriveType.VEHICLE:
                        self.spawn_vehicle(scenario_id, track)
                    elif track["type"] == MetaDriveType.CYCLIST:
                        self.spawn_cyclist(scenario_id, track)
                    elif track["type"] == MetaDriveType.PEDESTRIAN:
                        self.spawn_pedestrian(scenario_id, track)
                    elif track["type"] in [MetaDriveType.TRAFFIC_CONE, MetaDriveType.TRAFFIC_BARRIER]:
                        cls = TrafficBarrier if track["type"] == MetaDriveType.TRAFFIC_BARRIER else TrafficCone
                        self.spawn_static_object(cls, scenario_id, track)
                    else:
                        logger.info("Do not support {}".format(track["type"]))
                elif self.has_policy(self._scenario_id_to_obj_id[scenario_id], ReplayTrafficParticipantPolicy):
                    # static object will not be cleaned!
                    policy = self.get_policy(self._scenario_id_to_obj_id[scenario_id])
                    if policy.is_current_step_valid:
                        policy.act()
                    else:
                        self._obj_to_clean_this_frame.append(scenario_id)
        else:
            replay_done = True
            # clean replay vehicle
            for scenario_id, obj_id in self._scenario_id_to_obj_id.items():
                if self.has_policy(obj_id, ReplayTrafficParticipantPolicy) and not self.is_static_object(obj_id):
                    self._obj_to_clean_this_frame.append(scenario_id)

        for scenario_id in list(self._obj_to_clean_this_frame):
            obj_id = self._scenario_id_to_obj_id.pop(scenario_id)
            _scenario_id = self._obj_id_to_scenario_id.pop(obj_id)
            assert _scenario_id == scenario_id
            self.clear_objects([obj_id])

        return dict(default_agent=dict(replay_done=replay_done))

    @property
    def current_traffic_data(self):
        data = self.engine.data_manager.current_scenario["tracks"]
        return data

    @property
    def sdc_track_index(self):
        return str(self.engine.data_manager.current_scenario[SD.METADATA][SD.SDC_ID])

    @property
    def sdc_scenario_id(self):
        return self.sdc_track_index

    @property
    def sdc_object_id(self):
        return self.engine.agent_manager.agent_to_object(DEFAULT_AGENT)

    @property
    def current_scenario_length(self):
        return self.engine.data_manager.current_scenario_length

    def spawn_vehicle(self, v_id, track):
        state = parse_object_state(track, self.episode_step)

        # for each vehicle, we would like to know if it is static
        if v_id not in self._static_car_id and v_id not in self._moving_car_id:
            valid_points = track["state"]["position"][np.where(track["state"]["valid"])]
            moving = np.max(np.std(valid_points, axis=0)[:2]) > self.STATIC_THRESHOLD
            set_to_add = self._moving_car_id if moving else self._static_car_id
            set_to_add.add(v_id)

        # don't create in these two conditions
        if not state["valid"] or (self.engine.global_config["no_static_vehicles"] and v_id in self._static_car_id):
            return

        # if collision don't generate, unless ego car is in replay mode
        ego_pos = self.ego_vehicle.position
        heading_dist, side_dist = self.ego_vehicle.convert_to_local_coordinates(state["position"], ego_pos)
        if not self.is_ego_vehicle_replay and self._filter_overlapping_car and \
                abs(heading_dist) < self.GENERATION_FORWARD_CONSTRAINT and \
                abs(side_dist) < self.GENERATION_SIDE_CONSTRAINT:
            return

        # create vehicle
        if state["vehicle_class"]:
            vehicle_class = state["vehicle_class"]
        else:
            vehicle_class = get_vehicle_type(
                float(state["length"]), None if self.even_sample_v else self.np_random, self.need_default_vehicle
            )
        obj_name = v_id if self.engine.global_config["force_reuse_object_name"] else None
        v = self.spawn_object(
            vehicle_class,
            position=state["position"],
            heading=state["heading"],
            vehicle_config=self._traffic_v_config,
            name=obj_name
        )
        self._scenario_id_to_obj_id[v_id] = v.name
        self._obj_id_to_scenario_id[v.name] = v_id

        # add policy
        start_index, end_index = get_max_valid_indicis(track, self.episode_step)  # real end_index is end_index-1
        moving = track["state"]["position"][start_index][..., :2] - track["state"]["position"][end_index - 1][..., :2]
        length_ok = norm(moving[0], moving[1]) > self.IDM_CREATE_MIN_LENGTH
        heading_ok = abs(wrap_to_pi(self.ego_vehicle.heading_theta - state["heading"])) < np.pi / 2
        idm_ok = heading_dist < self.IDM_CREATE_FORWARD_CONSTRAINT and abs(
            side_dist
        ) < self.IDM_CREATE_SIDE_CONSTRAINT and heading_ok
        need_reactive_traffic = self.engine.global_config["reactive_traffic"]
        if not need_reactive_traffic or v_id in self._static_car_id or not idm_ok or not length_ok:
            policy = self.add_policy(v.name, ReplayTrafficParticipantPolicy, v, track)
            policy.act()
        else:
            idm_route = get_idm_route(track["state"]["position"][start_index:end_index][..., :2])
            # only not static and behind ego car, it can get reactive policy
            self.add_policy(
                v.name, ScenarioIDMPolicy, v, self.generate_seed(), idm_route,
                self.idm_policy_count % self.IDM_ACT_BATCH_SIZE
            )
            # no act() is required for IDMPolicy
            self.idm_policy_count += 1

    def spawn_pedestrian(self, scenario_id, track):
        state = parse_object_state(track, self.episode_step)
        if not state["valid"]:
            return
        obj_name = scenario_id if self.engine.global_config["force_reuse_object_name"] else None
        obj = self.spawn_object(
            Pedestrian,
            name=obj_name,
            position=state["position"],
            heading_theta=state["heading"],
        )
        self._scenario_id_to_obj_id[scenario_id] = obj.name
        self._obj_id_to_scenario_id[obj.name] = scenario_id
        policy = self.add_policy(obj.name, ReplayTrafficParticipantPolicy, obj, track)
        policy.act()

    def spawn_cyclist(self, scenario_id, track):
        state = parse_object_state(track, self.episode_step)
        if not state["valid"]:
            return
        obj_name = scenario_id if self.engine.global_config["force_reuse_object_name"] else None
        obj = self.spawn_object(
            Cyclist,
            name=obj_name,
            position=state["position"],
            heading_theta=state["heading"],
        )
        self._scenario_id_to_obj_id[scenario_id] = obj.name
        self._obj_id_to_scenario_id[obj.name] = scenario_id
        policy = self.add_policy(obj.name, ReplayTrafficParticipantPolicy, obj, track)
        policy.act()

    def spawn_static_object(self, cls, scenario_id, track):
        # some
        if scenario_id not in self._noise_object_id and scenario_id not in self._non_noise_object_id:
            valid_length = np.sum(track["state"]["valid"])
            set_to_add = self._noise_object_id if valid_length < self.MIN_VALID_FRAME_LEN else self._non_noise_object_id
            set_to_add.add(scenario_id)
        if scenario_id in self._noise_object_id:
            return

        state = parse_object_state(track, self.episode_step)
        if not state["valid"]:
            return
        obj_name = scenario_id if self.engine.global_config["force_reuse_object_name"] else None
        obj = self.spawn_object(
            cls,
            name=obj_name,
            position=state["position"],
            heading_theta=state["heading"],
            static=True if self.engine.global_config["static_traffic_object"] else False,
        )
        self._scenario_id_to_obj_id[scenario_id] = obj.name
        self._obj_id_to_scenario_id[obj.name] = scenario_id

    def get_state(self):
        # Record mapping from original_id to new_id
        ret = {}
        ret[SD.ORIGINAL_ID_TO_OBJ_ID] = self.scenario_id_to_obj_id
        ret[SD.OBJ_ID_TO_ORIGINAL_ID] = self.obj_id_to_scenario_id
        return ret

    @property
    def ego_vehicle(self):
        return self.engine.agents[DEFAULT_AGENT]

    def is_static_object(self, obj_id):
        return isinstance(self.spawned_objects[obj_id], TrafficBarrier) \
               or isinstance(self.spawned_objects[obj_id], TrafficCone)

    @property
    def obj_id_to_scenario_id(self):
        # For outside access, we return traffic vehicles and ego car
        ret = copy.copy(self._obj_id_to_scenario_id)
        ret[self.sdc_object_id] = self.sdc_scenario_id
        return ret

    @property
    def scenario_id_to_obj_id(self):
        # For outside access, we return traffic vehicles and ego car
        ret = copy.copy(self._scenario_id_to_obj_id)
        ret[self.sdc_scenario_id] = self.sdc_object_id
        return ret

    def get_traffic_v_config(self):
        v_config = copy.deepcopy(self.engine.global_config["vehicle_config"])
        v_config["need_navigation"] = False
        v_config.update(
            dict(
                show_navi_mark=False,
                show_dest_mark=False,
                enable_reverse=False,
                show_lidar=False,
                show_lane_line_detector=False,
                show_side_detector=False,
            )
        )
        return v_config



import math

from metadrive.envs.metadrive_env import MetaDriveEnv
from metadrive.utils import clip


def test_out_of_road():
    # env = MetaDriveEnv(dict(vehicle_config=dict(side_detector=dict(num_lasers=8))))
    for steering in [-0.01, 0.01]:
        for distance in [10, 50, 100]:
            env = MetaDriveEnv(
                dict(
                    map="SSSSSSSSSSS",
                    vehicle_config=dict(side_detector=dict(num_lasers=120, distance=distance)),
                    use_render=False
                )
            )
            try:
                obs, _ = env.reset()
                tolerance = math.sqrt(env.vehicle.WIDTH**2 + env.vehicle.LENGTH**2) / distance
                for _ in range(100000000):
                    o, r, tm, tc, i = env.step([steering, 1])
                    if tm or tc:
                        points = \
                            env.vehicle.side_detector.perceive(env.vehicle,
                                                               env.vehicle.engine.physics_world.static_world).cloud_points
                        assert min(points) < tolerance, (min(points), tolerance)
                        print(
                            "Side detector minimal distance: {}, Current distance: {}, steering: {}".format(
                                min(points) * distance, distance, steering
                            )
                        )
                        break
            finally:
                env.close()


def useless_left_right_distance_printing():
    # env = MetaDriveEnv(dict(vehicle_config=dict(side_detector=dict(num_lasers=8))))
    for steering in [-0.01, 0.01, -1, 1]:
        # for distance in [10, 50, 100]:
        env = MetaDriveEnv(
            dict(
                map="SSSSSSSSSSS", vehicle_config=dict(side_detector=dict(num_lasers=0, distance=50)), use_render=False
            )
        )
        try:
            for _ in range(100000000):
                o, r, tm, tc, i = env.step([steering, 1])
                vehicle = env.vehicle
                l, r = vehicle.dist_to_left_side, vehicle.dist_to_right_side
                total_width = float(
                    (vehicle.navigation.get_current_lane_num() + 1) * vehicle.navigation.get_current_lane_width()
                )
                print(
                    "Left {}, Right {}, Total {}. Clip Total {}".format(
                        l / total_width, r / total_width, (l + r) / total_width,
                        clip(l / total_width, 0, 1) + clip(r / total_width, 0, 1)
                    )
                )
                if d:
                    break
        finally:
            env.close()


if __name__ == '__main__':
    # test_out_of_road()
    useless_left_right_distance_printing()



from metadrive.envs.metadrive_env import MetaDriveEnv


def test_collision_with_vehicle(use_render=False):
    if not use_render:
        env = MetaDriveEnv({"traffic_density": 1.0, "map": "SSS"})
    else:
        env = MetaDriveEnv({"traffic_density": 1.0, "map": "SSS", "use_render": True})
    o, _ = env.reset()
    pass_test = False
    try:
        for i in range(1, 500):
            o, r, tm, tc, info = env.step([0, 1])
            if env.vehicle.crash_vehicle:
                pass_test = True
                break
        assert pass_test, "Collision function is broken!"
    finally:
        env.close()


def test_collision_with_sidewalk():
    env = MetaDriveEnv({"traffic_density": .0})
    o, _ = env.reset()
    pass_test = False
    try:
        for i in range(1, 100):
            o, r, tm, tc, info = env.step([-0.5, 1])
            if env.vehicle.crash_sidewalk:
                pass_test = True
                break
        assert pass_test, "Collision function is broken!"
    finally:
        env.close()


def test_line_contact():
    env = MetaDriveEnv({"traffic_density": .0})
    o, _ = env.reset()
    on_broken_line = False
    on_continuous_line = False
    try:
        for i in range(1, 100):
            o, r, tm, tc, info = env.step([-0.5, 1])
            on_broken_line = on_broken_line or env.vehicle.on_broken_line
            on_continuous_line = on_continuous_line or env.vehicle.on_white_continuous_line
        assert on_broken_line and on_continuous_line, "Collision function is broken!"
    finally:
        env.close()


if __name__ == '__main__':
    test_collision_with_sidewalk()
    # test_collision_with_vehicle(True)



import copy
from metadrive.type import MetaDriveType
from metadrive.constants import PGLineType, PGLineColor
from typing import List

from panda3d.core import NodePath

from metadrive.component.algorithm.BIG import BigGenerateMethod, BIG
from metadrive.component.map.base_map import BaseMap
from metadrive.component.pgblock.first_block import FirstPGBlock
from metadrive.component.road_network.node_road_network import NodeRoadNetwork
from metadrive.constants import Decoration
from metadrive.engine.core.physics_world import PhysicsWorld
from metadrive.utils import Config


def parse_map_config(easy_map_config, new_map_config, default_config):
    assert isinstance(new_map_config, Config)
    assert isinstance(default_config, Config)

    # Return the user specified config if overwritten
    if not default_config["map_config"].is_identical(new_map_config):
        new_map_config = default_config["map_config"].copy(unchangeable=False).update(new_map_config)
        assert default_config["map"] == easy_map_config
        return new_map_config

    if isinstance(easy_map_config, int):
        new_map_config[BaseMap.GENERATE_TYPE] = BigGenerateMethod.BLOCK_NUM
    elif isinstance(easy_map_config, str):
        new_map_config[BaseMap.GENERATE_TYPE] = BigGenerateMethod.BLOCK_SEQUENCE
    else:
        raise ValueError(
            "Unkown easy map config: {} and original map config: {}".format(easy_map_config, new_map_config)
        )
    new_map_config[BaseMap.GENERATE_CONFIG] = easy_map_config
    return new_map_config


class MapGenerateMethod:
    BIG_BLOCK_NUM = BigGenerateMethod.BLOCK_NUM
    BIG_BLOCK_SEQUENCE = BigGenerateMethod.BLOCK_SEQUENCE
    BIG_SINGLE_BLOCK = BigGenerateMethod.SINGLE_BLOCK
    PG_MAP_FILE = "pg_map_file"


class PGMap(BaseMap):
    def _generate(self):
        """
        We can override this function to introduce other methods!
        """
        parent_node_path, physics_world = self.engine.worldNP, self.engine.physics_world
        generate_type = self._config[self.GENERATE_TYPE]
        if generate_type == BigGenerateMethod.BLOCK_NUM or generate_type == BigGenerateMethod.BLOCK_SEQUENCE:
            self._big_generate(parent_node_path, physics_world)

        elif generate_type == MapGenerateMethod.PG_MAP_FILE:
            # other config such as lane width, num and seed will be valid, since they will be read from file
            blocks_config = self._config[self.GENERATE_CONFIG]
            self._config_generate(blocks_config, parent_node_path, physics_world)
        else:
            raise ValueError("Map can not be created by {}".format(generate_type))
        self.road_network.after_init()

    def _big_generate(self, parent_node_path: NodePath, physics_world: PhysicsWorld):
        big_map = BIG(
            self._config[self.LANE_NUM],
            self._config[self.LANE_WIDTH],
            self.road_network,
            parent_node_path,
            physics_world,
            # self._config["block_type_version"],
            exit_length=self._config["exit_length"],
            random_seed=self.engine.global_random_seed,
            block_dist_config=self.engine.global_config["block_dist_config"]
        )
        big_map.generate(self._config[self.GENERATE_TYPE], self._config[self.GENERATE_CONFIG])
        self.blocks = big_map.blocks
        del big_map

    def _config_generate(self, blocks_config: List, parent_node_path: NodePath, physics_world: PhysicsWorld):
        assert len(self.road_network.graph) == 0, "These Map is not empty, please create a new map to read config"
        last_block = FirstPGBlock(
            global_network=self.road_network,
            lane_width=self._config[self.LANE_WIDTH],
            lane_num=self._config[self.LANE_NUM],
            render_root_np=parent_node_path,
            physics_world=physics_world,
            length=self._config["exit_length"],
            ignore_intersection_checking=True
        )
        self.blocks.append(last_block)
        for block_index, b in enumerate(blocks_config[1:], 1):
            block_type = self.engine.global_config["block_dist_config"].get_block(b.pop(self.BLOCK_ID))
            pre_block_socket_index = b.pop(self.PRE_BLOCK_SOCKET_INDEX)
            last_block = block_type(
                block_index,
                last_block.get_socket(pre_block_socket_index),
                self.road_network,
                random_seed=self.engine.global_random_seed,
                ignore_intersection_checking=True
            )
            last_block.construct_from_config(b, parent_node_path, physics_world)
            self.blocks.append(last_block)

    @property
    def road_network_type(self):
        return NodeRoadNetwork

    def get_meta_data(self):
        assert self.blocks is not None and len(self.blocks) > 0, "Please generate Map before saving it"
        map_config = []
        for b in self.blocks:
            b_config = b.get_config()
            json_config = b_config.get_serializable_dict()
            json_config[self.BLOCK_ID] = b.ID
            json_config[self.PRE_BLOCK_SOCKET_INDEX] = b.pre_block_socket_index
            map_config.append(json_config)

        saved_data = copy.deepcopy({self.BLOCK_SEQUENCE: map_config, "map_config": self.config.copy()})
        saved_data.update(super(PGMap, self).get_meta_data())
        return saved_data

    def show_coordinates(self):
        lanes = []
        for to_ in self.road_network.graph.values():
            for lanes_to_add in to_.values():
                lanes += lanes_to_add
        self.engine.show_lane_coordinates(lanes)

    def get_boundary_line_vector(self, interval):
        map = self
        ret = {}
        for _from in map.road_network.graph.keys():
            decoration = True if _from == Decoration.start else False
            for _to in map.road_network.graph[_from].keys():
                for l in map.road_network.graph[_from][_to]:
                    sides = 2 if l is map.road_network.graph[_from][_to][-1] or decoration else 1
                    for side in range(sides):
                        type = l.line_types[side]
                        if type == PGLineType.NONE:
                            continue
                        color = l.line_colors[side]
                        line_type = self.get_line_type(type, color)
                        lateral = l.width_at(0) / 2
                        if side == 0:
                            lateral *= -1
                        ret["{}_{}".format(l.index, side)] = {
                            "type": line_type,
                            "polyline": l.get_polyline(interval, lateral),
                            "speed_limit_kmh": l.speed_limit
                        }
        return ret

    def get_line_type(self, type, color):
        if type == PGLineType.CONTINUOUS and color == PGLineColor.YELLOW:
            return MetaDriveType.LINE_SOLID_SINGLE_YELLOW
        elif type == PGLineType.BROKEN and color == PGLineColor.YELLOW:
            return MetaDriveType.LINE_BROKEN_SINGLE_YELLOW
        elif type == PGLineType.CONTINUOUS and color == PGLineColor.GREY:
            return MetaDriveType.LINE_SOLID_SINGLE_WHITE
        elif type == PGLineType.BROKEN and color == PGLineColor.GREY:
            return MetaDriveType.LINE_BROKEN_SINGLE_WHITE
        elif type == PGLineType.SIDE:
            return MetaDriveType.LINE_SOLID_SINGLE_WHITE
        else:
            # Unknown line type
            return MetaDriveType.LINE_UNKNOWN



import logging
import os
import tempfile
from os.path import join
import hydra
import nuplan.planning.script.builders.worker_pool_builder
from nuplan.planning.script.builders.scenario_building_builder import build_scenario_builder
from nuplan.planning.script.builders.scenario_filter_builder import build_scenario_filter
from nuplan.planning.script.utils import set_up_common_builder
from dataclasses import dataclass
from metadrive.manager.base_manager import BaseManager
from metadrive.utils.utils import is_win

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)

ROOT = os.path.dirname(__file__)


# Copied from nuplan-devkit tutorial_utils.py
@dataclass
class HydraConfigPaths:
    """
    Stores relative hydra paths to declutter tutorial.
    """

    common_dir: str
    config_name: str
    config_path: str
    experiment_dir: str


def construct_simulation_hydra_paths(base_config_path: str):
    """
    Specifies relative paths to simulation configs to pass to hydra to declutter tutorial.
    :param base_config_path: Base config path.
    :return: Hydra config path.
    """
    common_dir = "file://" + join(base_config_path, 'config', 'common')
    config_name = 'default_simulation'
    config_path = join(base_config_path, 'config', 'simulation')
    experiment_dir = "file://" + join(base_config_path, 'experiments')
    return HydraConfigPaths(common_dir, config_name, config_path, experiment_dir)


class NuPlanDataManager(BaseManager):
    """
    This manager serves as the interface between nuplan-devkit and MetaDrive
    """
    NUPLAN_PACKAGE_PATH = os.path.dirname(nuplan.__file__)
    NUPLAN_ROOT_PATH = os.path.dirname(NUPLAN_PACKAGE_PATH)
    NUPLAN_TUTORIAL_PATH = os.path.join(os.path.dirname(nuplan.__file__), "tutorial")

    def __init__(self):
        super(NuPlanDataManager, self).__init__()
        logger.info("\n \n ############### Start Loading NuPlan Data ############### \n")
        # get original nuplan cfg
        cfg = self._get_nuplan_cfg()
        profiler_name = 'building_simulation'
        common_builder = set_up_common_builder(cfg=cfg, profiler_name=profiler_name)

        # Build scenario builder
        scenario_builder = build_scenario_builder(cfg=cfg)
        scenario_filter = build_scenario_filter(cfg.scenario_filter)

        # get scenarios from database
        self._nuplan_scenarios = scenario_builder.get_scenarios(scenario_filter, common_builder.worker)

        # filter scenario according to config
        self.start_scenario_index = self.engine.global_config["start_scenario_index"]
        self._num_scenarios = self.engine.global_config["num_scenarios"]
        assert len(self._nuplan_scenarios) >= self.start_scenario_index + self.num_scenarios, \
            "Number of scenes are not enough, " \
            "\n num nuplan scenarios: {}" \
            "\n start_scenario_index: {}" \
            "\n scenario num: {}".format(len(self._nuplan_scenarios), self.start_scenario_index, self.num_scenarios)
        logger.info("\n \n ############### Finish Loading NuPlan Data ############### \n")

        self._current_scenario_index = None

    @property
    def time_interval(self):
        return self.current_scenario.database_interval

    @property
    def num_scenarios(self):
        return self._num_scenarios

    @property
    def current_scenario_index(self):
        return self._current_scenario_index

    @property
    def current_scenario(self):
        return self._nuplan_scenarios[self._current_scenario_index]

    def _get_nuplan_cfg(self):
        BASE_CONFIG_PATH = os.path.join(self.NUPLAN_PACKAGE_PATH, "planning", "script")
        simulation_hydra_paths = construct_simulation_hydra_paths(BASE_CONFIG_PATH)

        # Initialize configuration management system
        hydra.core.global_hydra.GlobalHydra.instance().clear()  # reinitialize hydra if already initialized
        hydra.initialize_config_dir(config_dir=simulation_hydra_paths.config_path)

        SAVE_DIR = tempfile.mkdtemp()
        EGO_CONTROLLER = 'perfect_tracking_controller'  # [log_play_back_controller, perfect_tracking_controller]
        OBSERVATION = 'box_observation'  # [box_observation, idm_agents_observation, lidar_pc_observation]

        DATASET_PARAMS = self.engine.global_config["DATASET_PARAMS"]

        # Compose the configuration
        overrides = [
            f'group={SAVE_DIR}',
            'worker=sequential',
            f'ego_controller={EGO_CONTROLLER}',
            f'observation={OBSERVATION}',
            f'hydra.searchpath=[{simulation_hydra_paths.common_dir}, {simulation_hydra_paths.experiment_dir}]',
            'output_dir=${group}/${experiment}',
            *DATASET_PARAMS,
        ]
        if is_win():
            overrides.extend(
                [
                    f'job_name=planner_tutorial',
                    'experiment=${experiment_name}/${job_name}/${experiment_time}',
                ]
            )
        else:
            overrides.append(f'experiment_name=planner_tutorial')
        cfg = hydra.compose(config_name=simulation_hydra_paths.config_name, overrides=overrides)
        return cfg

    def get_scenario(self, index, force_get_current_scenario=True):
        assert self.start_scenario_index <= index < self.start_scenario_index + self.num_scenarios
        if force_get_current_scenario:
            assert index == self.random_seed
            return self.current_scenario
        else:
            return self._nuplan_scenarios[index]

    def seed(self, random_seed):
        assert self.start_scenario_index <= random_seed < self.start_scenario_index + self.num_scenarios
        super(NuPlanDataManager, self).seed(random_seed)
        self._current_scenario_index = random_seed

    @property
    def current_scenario_length(self):
        return self.current_scenario.get_number_of_iterations()

    def get_metadata(self):
        state = super(NuPlanDataManager, self).get_metadata()
        raw_data = self.current_scenario
        state["raw_data"] = raw_data
        return state



from metadrive.component.pgblock.t_intersection import TInterSection
from metadrive.component.pg_space import Parameter


class StdTInterSection(TInterSection):
    def _try_plug_into_previous_block(self) -> bool:
        self._config[Parameter.change_lane_num] = 0
        success = super(StdTInterSection, self)._try_plug_into_previous_block()
        return success



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'main_window.ui'
#
# Created by: PyQt5 UI code generator 5.7
#
# WARNING! All changes made in this file will be lost!

from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_MainWindow(object):
    def setupUi(self, MainWindow):
        MainWindow.setObjectName("MainWindow")
        MainWindow.resize(262, 608)
        MainWindow.setMinimumSize(QtCore.QSize(262, 608))
        MainWindow.setMaximumSize(QtCore.QSize(262, 5560))
        MainWindow.setStyleSheet(
            "QMainWindow { background: #fff;}\n"
            "*, QLabel { font-family: Roboto; font-weight: 300; }\n"
            "\n"
            "QSlider::groove:horizontal {\n"
            "    border: 0;\n"
            "    height: 4px;\n"
            "    background: rgba(0, 0, 0, 15);\n"
            "}\n"
            "\n"
            "QSlider::handle:horizonal {\n"
            "    background: #666;\n"
            "    width: 8px;\n"
            "    margin: -6px 0;\n"
            "}\n"
            "\n"
            "QSlider::handle:horizontal:disabled {\n"
            "    background: #ccc;\n"
            "}\n"
            "QSlider::groove:horizontal:disabled {\n"
            "    background: rgba(0, 0, 0, 8);\n"
            "}\n"
            "\n"
            "QRadioButton {\n"
            "    background: #fafafa;\n"
            "    padding: 4px 6px 3px;\n"
            "   font-size: 12px;\n"
            "    color: #333;\n"
            "}\n"
            "\n"
            "QRadioButton::indicator {\n"
            "    width: 15px;\n"
            "    height: 15px;\n"
            "    margin-top: -1px;\n"
            "}\n"
            "\n"
            "QRadioButton:checked {\n"
            "background: #eee;\n"
            "}\n"
            "\n"
            "QRadioButton:hover {\n"
            "background: #eaeaea;\n"
            "}\n"
            "\n"
            "QComboBox {\n"
            "border: 0;\n"
            "background: #fff;\n"
            "padding: 7px 10px 3px;\n"
            "font-family: Roboto;\n"
            "color: #777;\n"
            "font-size: 12px;\n"
            "}\n"
            "\n"
            "QComboBox::down-arrow {\n"
            "    border: 0;\n"
            "    margin: -3px;\n"
            "    background: #ddd;\n"
            "    margin-left: -16px;\n"
            "margin-bottom: -4px;\n"
            "    image: url(\":/img/down-arrow.png\");\n"
            "}\n"
            "\n"
            "QCheckBox::indicator {\n"
            "    border: 1px solid #aaa;\n"
            "    background: #39f;\n"
            "    width: 14px;\n"
            "    height: 14px;\n"
            "    image: url(\":/img/checked.png\");\n"
            "}\n"
            "\n"
            "QCheckBox {\n"
            "text-transform: uppercase;\n"
            "color: #555;\n"
            "}\n"
            "\n"
            "QCheckBox::indicator:unchecked {\n"
            "    background: #fff;\n"
            "    image: none;\n"
            "}"
        )
        self.centralwidget = QtWidgets.QWidget(MainWindow)
        self.centralwidget.setObjectName("centralwidget")
        self.verticalLayout = QtWidgets.QVBoxLayout(self.centralwidget)
        self.verticalLayout.setContentsMargins(0, 0, 0, 0)
        self.verticalLayout.setSpacing(0)
        self.verticalLayout.setObjectName("verticalLayout")
        self.frame = QtWidgets.QFrame(self.centralwidget)
        self.frame.setMinimumSize(QtCore.QSize(0, 65))
        self.frame.setMaximumSize(QtCore.QSize(16777215, 65))
        self.frame.setFrameShape(QtWidgets.QFrame.StyledPanel)
        self.frame.setFrameShadow(QtWidgets.QFrame.Raised)
        self.frame.setObjectName("frame")
        self.label = QtWidgets.QLabel(self.frame)
        self.label.setGeometry(QtCore.QRect(20, 26, 261, 31))
        self.label.setStyleSheet("font-size: 20px; color: #03C03C;\n" "font-weight: 100;")
        self.label.setObjectName("label")
        self.label_9 = QtWidgets.QLabel(self.frame)
        self.label_9.setGeometry(QtCore.QRect(20, 6, 261, 31))
        self.label_9.setStyleSheet("font-size: 13px;\n" "color: #aaa;\n" "font-weight: 600;")
        self.label_9.setObjectName("label_9")
        self.verticalLayout.addWidget(self.frame)
        self.horizontalLayout = QtWidgets.QHBoxLayout()
        self.horizontalLayout.setSpacing(0)
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.sidebar_material = QtWidgets.QFrame(self.centralwidget)
        self.sidebar_material.setMinimumSize(QtCore.QSize(0, 0))
        self.sidebar_material.setMaximumSize(QtCore.QSize(999999, 16777215))
        self.sidebar_material.setStyleSheet(
            "QFrame { background: #eee; } QFrame>QLabel { text-transform: uppercase; color: #888;  background: transparent; }"
        )
        self.sidebar_material.setFrameShape(QtWidgets.QFrame.StyledPanel)
        self.sidebar_material.setFrameShadow(QtWidgets.QFrame.Raised)
        self.sidebar_material.setObjectName("sidebar_material")
        self.label_2 = QtWidgets.QLabel(self.sidebar_material)
        self.label_2.setGeometry(QtCore.QRect(20, 20, 61, 16))
        self.label_2.setObjectName("label_2")
        self.cb_material = QtWidgets.QComboBox(self.sidebar_material)
        self.cb_material.setGeometry(QtCore.QRect(20, 40, 221, 30))
        self.cb_material.setObjectName("cb_material")
        self.cb_material.addItem("")
        self.label_14 = QtWidgets.QLabel(self.sidebar_material)
        self.label_14.setGeometry(QtCore.QRect(20, 80, 141, 16))
        self.label_14.setObjectName("label_14")
        self.cb_shading_model = QtWidgets.QComboBox(self.sidebar_material)
        self.cb_shading_model.setGeometry(QtCore.QRect(20, 100, 221, 30))
        self.cb_shading_model.setObjectName("cb_shading_model")
        self.cb_shading_model.addItem("")
        self.cb_shading_model.addItem("")
        self.cb_shading_model.addItem("")
        self.cb_metallic = QtWidgets.QCheckBox(self.sidebar_material)
        self.cb_metallic.setGeometry(QtCore.QRect(20, 140, 81, 17))
        self.cb_metallic.setObjectName("cb_metallic")
        self.horizontalLayoutWidget_3 = QtWidgets.QWidget(self.sidebar_material)
        self.horizontalLayoutWidget_3.setGeometry(QtCore.QRect(30, 240, 201, 73))
        self.horizontalLayoutWidget_3.setObjectName("horizontalLayoutWidget_3")
        self.horizontalLayout_4 = QtWidgets.QHBoxLayout(self.horizontalLayoutWidget_3)
        self.horizontalLayout_4.setContentsMargins(0, 0, 0, 0)
        self.horizontalLayout_4.setSpacing(6)
        self.horizontalLayout_4.setObjectName("horizontalLayout_4")
        self.gridLayout = QtWidgets.QGridLayout()
        self.gridLayout.setObjectName("gridLayout")
        self.lbl_basecolor_val2 = QtWidgets.QLabel(self.horizontalLayoutWidget_3)
        self.lbl_basecolor_val2.setMinimumSize(QtCore.QSize(30, 0))
        self.lbl_basecolor_val2.setMaximumSize(QtCore.QSize(30, 16777215))
        self.lbl_basecolor_val2.setStyleSheet("background: transparent;")
        self.lbl_basecolor_val2.setAlignment(QtCore.Qt.AlignCenter)
        self.lbl_basecolor_val2.setObjectName("lbl_basecolor_val2")
        self.gridLayout.addWidget(self.lbl_basecolor_val2, 1, 2, 1, 1)
        self.basecolor_3 = QtWidgets.QSlider(self.horizontalLayoutWidget_3)
        self.basecolor_3.setMaximum(100)
        self.basecolor_3.setProperty("value", 53)
        self.basecolor_3.setOrientation(QtCore.Qt.Horizontal)
        self.basecolor_3.setObjectName("basecolor_3")
        self.gridLayout.addWidget(self.basecolor_3, 2, 1, 1, 1)
        self.basecolor_1 = QtWidgets.QSlider(self.horizontalLayoutWidget_3)
        self.basecolor_1.setMaximum(100)
        self.basecolor_1.setProperty("value", 12)
        self.basecolor_1.setOrientation(QtCore.Qt.Horizontal)
        self.basecolor_1.setObjectName("basecolor_1")
        self.gridLayout.addWidget(self.basecolor_1, 0, 1, 1, 1)
        self.basecolor_2 = QtWidgets.QSlider(self.horizontalLayoutWidget_3)
        self.basecolor_2.setMaximum(100)
        self.basecolor_2.setProperty("value", 43)
        self.basecolor_2.setOrientation(QtCore.Qt.Horizontal)
        self.basecolor_2.setObjectName("basecolor_2")
        self.gridLayout.addWidget(self.basecolor_2, 1, 1, 1, 1)
        self.lbl_basecolor_val3 = QtWidgets.QLabel(self.horizontalLayoutWidget_3)
        self.lbl_basecolor_val3.setMinimumSize(QtCore.QSize(30, 0))
        self.lbl_basecolor_val3.setMaximumSize(QtCore.QSize(30, 16777215))
        self.lbl_basecolor_val3.setStyleSheet("background: transparent;")
        self.lbl_basecolor_val3.setAlignment(QtCore.Qt.AlignCenter)
        self.lbl_basecolor_val3.setObjectName("lbl_basecolor_val3")
        self.gridLayout.addWidget(self.lbl_basecolor_val3, 2, 2, 1, 1)
        self.lbl_basecolor_val1 = QtWidgets.QLabel(self.horizontalLayoutWidget_3)
        self.lbl_basecolor_val1.setMinimumSize(QtCore.QSize(30, 0))
        self.lbl_basecolor_val1.setMaximumSize(QtCore.QSize(30, 16777215))
        self.lbl_basecolor_val1.setStyleSheet("background: transparent;")
        self.lbl_basecolor_val1.setAlignment(QtCore.Qt.AlignCenter)
        self.lbl_basecolor_val1.setObjectName("lbl_basecolor_val1")
        self.gridLayout.addWidget(self.lbl_basecolor_val1, 0, 2, 1, 1)
        self.lbl_basecolor1 = QtWidgets.QLabel(self.horizontalLayoutWidget_3)
        self.lbl_basecolor1.setMinimumSize(QtCore.QSize(11, 0))
        self.lbl_basecolor1.setStyleSheet("background: transparent; color: #555; ")
        self.lbl_basecolor1.setObjectName("lbl_basecolor1")
        self.gridLayout.addWidget(self.lbl_basecolor1, 0, 0, 1, 1)
        self.lbl_basecolor2 = QtWidgets.QLabel(self.horizontalLayoutWidget_3)
        self.lbl_basecolor2.setMinimumSize(QtCore.QSize(11, 0))
        self.lbl_basecolor2.setStyleSheet("background: transparent; color: #555; ")
        self.lbl_basecolor2.setObjectName("lbl_basecolor2")
        self.gridLayout.addWidget(self.lbl_basecolor2, 1, 0, 1, 1)
        self.lbl_basecolor3 = QtWidgets.QLabel(self.horizontalLayoutWidget_3)
        self.lbl_basecolor3.setMinimumSize(QtCore.QSize(11, 0))
        self.lbl_basecolor3.setStyleSheet("background: transparent; color: #555; ")
        self.lbl_basecolor3.setObjectName("lbl_basecolor3")
        self.gridLayout.addWidget(self.lbl_basecolor3, 2, 0, 1, 1)
        self.horizontalLayout_4.addLayout(self.gridLayout)
        self.lbl_color_preview = QtWidgets.QLabel(self.horizontalLayoutWidget_3)
        self.lbl_color_preview.setMinimumSize(QtCore.QSize(30, 0))
        self.lbl_color_preview.setMaximumSize(QtCore.QSize(30, 16777215))
        self.lbl_color_preview.setStyleSheet("background: rgba(60, 120, 255);")
        self.lbl_color_preview.setText("")
        self.lbl_color_preview.setObjectName("lbl_color_preview")
        self.horizontalLayout_4.addWidget(self.lbl_color_preview)
        self.label_15 = QtWidgets.QLabel(self.sidebar_material)
        self.label_15.setGeometry(QtCore.QRect(20, 180, 131, 16))
        self.label_15.setObjectName("label_15")
        self.label_16 = QtWidgets.QLabel(self.sidebar_material)
        self.label_16.setGeometry(QtCore.QRect(20, 340, 131, 16))
        self.label_16.setObjectName("label_16")
        self.slider_roughness = QtWidgets.QSlider(self.sidebar_material)
        self.slider_roughness.setGeometry(QtCore.QRect(20, 360, 191, 19))
        self.slider_roughness.setMaximum(100)
        self.slider_roughness.setProperty("value", 32)
        self.slider_roughness.setOrientation(QtCore.Qt.Horizontal)
        self.slider_roughness.setObjectName("slider_roughness")
        self.label_17 = QtWidgets.QLabel(self.sidebar_material)
        self.label_17.setGeometry(QtCore.QRect(20, 390, 131, 16))
        self.label_17.setObjectName("label_17")
        self.slider_specular = QtWidgets.QSlider(self.sidebar_material)
        self.slider_specular.setGeometry(QtCore.QRect(20, 410, 191, 19))
        self.slider_specular.setMaximum(100)
        self.slider_specular.setSliderPosition(20)
        self.slider_specular.setOrientation(QtCore.Qt.Horizontal)
        self.slider_specular.setObjectName("slider_specular")
        self.label_18 = QtWidgets.QLabel(self.sidebar_material)
        self.label_18.setGeometry(QtCore.QRect(20, 440, 131, 16))
        self.label_18.setObjectName("label_18")
        self.slider_normal = QtWidgets.QSlider(self.sidebar_material)
        self.slider_normal.setGeometry(QtCore.QRect(20, 460, 191, 19))
        self.slider_normal.setMaximum(100)
        self.slider_normal.setSliderPosition(80)
        self.slider_normal.setOrientation(QtCore.Qt.Horizontal)
        self.slider_normal.setObjectName("slider_normal")
        self.lbl_roughness = QtWidgets.QLabel(self.sidebar_material)
        self.lbl_roughness.setGeometry(QtCore.QRect(220, 360, 31, 20))
        self.lbl_roughness.setStyleSheet("color: #333;")
        self.lbl_roughness.setObjectName("lbl_roughness")
        self.lbl_specular = QtWidgets.QLabel(self.sidebar_material)
        self.lbl_specular.setGeometry(QtCore.QRect(220, 410, 31, 20))
        self.lbl_specular.setStyleSheet("color: #333;")
        self.lbl_specular.setObjectName("lbl_specular")
        self.lbl_normal = QtWidgets.QLabel(self.sidebar_material)
        self.lbl_normal.setGeometry(QtCore.QRect(220, 459, 31, 20))
        self.lbl_normal.setStyleSheet("color: #333;")
        self.lbl_normal.setObjectName("lbl_normal")
        self.frame_2 = QtWidgets.QFrame(self.sidebar_material)
        self.frame_2.setGeometry(QtCore.QRect(-10, 166, 311, 561))
        self.frame_2.setStyleSheet("QFrame { background: #fff; }")
        self.frame_2.setFrameShape(QtWidgets.QFrame.StyledPanel)
        self.frame_2.setFrameShadow(QtWidgets.QFrame.Raised)
        self.frame_2.setObjectName("frame_2")
        self.lbl_param1 = QtWidgets.QLabel(self.frame_2)
        self.lbl_param1.setGeometry(QtCore.QRect(230, 339, 31, 20))
        self.lbl_param1.setStyleSheet("color: #333;")
        self.lbl_param1.setObjectName("lbl_param1")
        self.slider_param1 = QtWidgets.QSlider(self.frame_2)
        self.slider_param1.setEnabled(False)
        self.slider_param1.setGeometry(QtCore.QRect(30, 340, 191, 19))
        self.slider_param1.setMaximum(100)
        self.slider_param1.setSliderPosition(80)
        self.slider_param1.setOrientation(QtCore.Qt.Horizontal)
        self.slider_param1.setObjectName("slider_param1")
        self.lbl_shading_model_param1 = QtWidgets.QLabel(self.frame_2)
        self.lbl_shading_model_param1.setGeometry(QtCore.QRect(30, 320, 141, 16))
        self.lbl_shading_model_param1.setObjectName("lbl_shading_model_param1")
        self.cb_rgb = QtWidgets.QRadioButton(self.frame_2)
        self.cb_rgb.setGeometry(QtCore.QRect(30, 34, 60, 27))
        self.cb_rgb.setChecked(True)
        self.cb_rgb.setObjectName("cb_rgb")
        self.cb_srgb = QtWidgets.QRadioButton(self.frame_2)
        self.cb_srgb.setGeometry(QtCore.QRect(92, 34, 71, 27))
        self.cb_srgb.setObjectName("cb_srgb")
        self.cb_hsv = QtWidgets.QRadioButton(self.frame_2)
        self.cb_hsv.setGeometry(QtCore.QRect(165, 34, 62, 27))
        self.cb_hsv.setObjectName("cb_hsv")
        self.frame_3 = QtWidgets.QFrame(self.frame_2)
        self.frame_3.setGeometry(QtCore.QRect(30, 59, 220, 96))
        self.frame_3.setStyleSheet("QFrame { background: #eee; }")
        self.frame_3.setFrameShape(QtWidgets.QFrame.StyledPanel)
        self.frame_3.setFrameShadow(QtWidgets.QFrame.Raised)
        self.frame_3.setObjectName("frame_3")
        self.lbl_param1.raise_()
        self.slider_param1.raise_()
        self.lbl_shading_model_param1.raise_()
        self.cb_rgb.raise_()
        self.cb_srgb.raise_()
        self.lbl_basecolor_val1.raise_()
        self.cb_hsv.raise_()
        self.frame_3.raise_()
        self.frame_2.raise_()
        self.label_2.raise_()
        self.cb_material.raise_()
        self.label_14.raise_()
        self.cb_shading_model.raise_()
        self.cb_metallic.raise_()
        self.horizontalLayoutWidget_3.raise_()
        self.label_15.raise_()
        self.label_16.raise_()
        self.slider_roughness.raise_()
        self.label_17.raise_()
        self.slider_specular.raise_()
        self.label_18.raise_()
        self.slider_normal.raise_()
        self.lbl_roughness.raise_()
        self.lbl_specular.raise_()
        self.lbl_normal.raise_()
        self.horizontalLayout.addWidget(self.sidebar_material)
        self.verticalLayout.addLayout(self.horizontalLayout)
        MainWindow.setCentralWidget(self.centralwidget)

        self.retranslateUi(MainWindow)
        QtCore.QMetaObject.connectSlotsByName(MainWindow)

    def retranslateUi(self, MainWindow):
        _translate = QtCore.QCoreApplication.translate
        MainWindow.setWindowTitle(_translate("MainWindow", "Render Pipeline Material Editor"))
        self.label.setText(_translate("MainWindow", "MATERIAL EDITOR"))
        self.label_9.setText(_translate("MainWindow", "RENDER PIPELINE"))
        self.label_2.setText(_translate("MainWindow", "Material"))
        self.cb_material.setItemText(0, _translate("MainWindow", "tree01"))
        self.label_14.setText(_translate("MainWindow", "Shading model"))
        self.cb_shading_model.setItemText(0, _translate("MainWindow", "Default"))
        self.cb_shading_model.setItemText(1, _translate("MainWindow", "Emissive"))
        self.cb_shading_model.setItemText(2, _translate("MainWindow", "Transparent"))
        self.cb_metallic.setText(_translate("MainWindow", "Metallic"))
        self.lbl_basecolor_val2.setText(_translate("MainWindow", "0.43"))
        self.lbl_basecolor_val3.setText(_translate("MainWindow", "0.53"))
        self.lbl_basecolor_val1.setText(_translate("MainWindow", "0.12"))
        self.lbl_basecolor1.setText(_translate("MainWindow", "R"))
        self.lbl_basecolor2.setText(_translate("MainWindow", "G"))
        self.lbl_basecolor3.setText(_translate("MainWindow", "B"))
        self.label_15.setText(_translate("MainWindow", "Basecolor"))
        self.label_16.setText(_translate("MainWindow", "Roughness"))
        self.label_17.setText(_translate("MainWindow", "Specular IOR"))
        self.label_18.setText(_translate("MainWindow", "Normal-Strength"))
        self.lbl_roughness.setText(_translate("MainWindow", "0.32"))
        self.lbl_specular.setText(_translate("MainWindow", "1.43"))
        self.lbl_normal.setText(_translate("MainWindow", "0.80"))
        self.lbl_param1.setText(_translate("MainWindow", "0.20"))
        self.lbl_shading_model_param1.setText(_translate("MainWindow", "Shading-model-param"))
        self.cb_rgb.setText(_translate("MainWindow", "RGB"))
        self.cb_srgb.setText(_translate("MainWindow", "SRGB"))
        self.cb_hsv.setText(_translate("MainWindow", "HSV"))


from . import resources_rc



"""

Render service to render previews of materials

"""

from __future__ import print_function

import sys
import socket
import time
import pickle

from threading import Thread

from panda3d.core import load_prc_file_data, Filename, Mat4
from panda3d.core import CS_zup_right, CS_yup_right, BamCache
from direct.showbase.ShowBase import ShowBase

sys.path.insert(0, "../../")
from rpcore import RenderPipeline, PointLight  # noqa


class Application(ShowBase):

    ICOMING_PORT = 62360

    def __init__(self):
        load_prc_file_data("", "win-size 512 512")
        load_prc_file_data("", "window-type offscreen")
        load_prc_file_data("", "model-cache-dir")
        load_prc_file_data("", "model-cache-textures #f")
        load_prc_file_data("", "textures-power-2 none")
        load_prc_file_data("", "alpha-bits 0")
        load_prc_file_data("", "print-pipe-types #f")

        # Construct render pipeline
        self.render_pipeline = RenderPipeline()
        self.render_pipeline.mount_mgr.config_dir = "config/"
        self.render_pipeline.create(self)

        self.setup_scene()

        # Disable model caching
        BamCache.get_global_ptr().cache_models = False

        self.update_queue = []
        self.start_listen()

        # Render initial frames
        for i in range(10):
            self.taskMgr.step()

        last_update = 0.0
        self.scene_node = None

        # Wait for updates
        while True:

            # Update once in a while
            curr_time = time.time()
            if curr_time > last_update + 1.0:
                last_update = curr_time
                self.taskMgr.step()

            if self.update_queue:
                if self.scene_node:
                    self.scene_node.remove_node()

                # Only take the latest packet
                payload = self.update_queue.pop(0)
                print("RENDERING:", payload)

                scene = self.loader.loadModel(Filename.from_os_specific(payload["scene"]))

                for light in scene.find_all_matches("**/+PointLight"):
                    light.remove_node()
                for light in scene.find_all_matches("**/+Spotlight"):
                    light.remove_node()

                # Find camera
                main_cam = scene.find("**/Camera")
                if main_cam:
                    transform_mat = main_cam.get_transform(self.render).get_mat()
                    transform_mat = Mat4.convert_mat(CS_zup_right, CS_yup_right) * transform_mat
                    self.camera.set_mat(transform_mat)
                else:
                    print("WARNING: No camera found")
                    self.camera.set_pos(0, -3.5, 0)
                    self.camera.look_at(0, -2.5, 0)

                self.camLens.set_fov(64.0)

                self.scene_node = scene
                scene.reparent_to(self.render)

                # Render scene
                for i in range(8):
                    self.taskMgr.step()

                dest_path = Filename.from_os_specific(payload["dest"])
                print("Saving screenshot to", dest_path)
                self.win.save_screenshot(dest_path)
                self.notify_about_finish(int(payload["pingback_port"]))

    def start_listen(self):
        """ Starts the listener thread """
        thread = Thread(target=self.listener_thread, args=(), name="ListenerThread")
        thread.setDaemon(True)
        thread.start()
        return thread

    def listener_thread(self):
        """ Thread which listens to incoming updates """
        sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        print("Listening on 127.0.0.1:" + str(self.ICOMING_PORT))
        try:
            sock.bind(("127.0.0.1", self.ICOMING_PORT))
            while True:
                data, addr = sock.recvfrom(8192)
                self.handle_data(data)
        except Exception as msg:
            print("Failed to bind to address! Reason:", msg)
        finally:
            sock.close()

    def handle_data(self, data):
        """ Handles a new update """
        # print("Got:", data)
        unpacked_data = pickle.loads(data)
        # print("Data = ", unpacked_data)
        self.update_queue.append(unpacked_data)

    def notify_about_finish(self, port):
        """ Notifies the caller that the result finished """
        print("Sending finish result to localhost:" + str(port))
        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        sock.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)
        try:
            sock.connect(("localhost", port))
        except Exception as msg:
            print("Could not send finish result: ", msg)
            return
        sock.sendall(b"done")
        print("Sent done flag.")
        sock.close()

    def setup_scene(self):
        """ Setups the basic scene geometry """
        self.disableMouse()
        self.render2d.hide()
        self.aspect2d.hide()

        light = PointLight()
        light.pos = 20.0, -0.85, -1.31
        light.radius = 100.0
        light.energy = 2500
        light.set_color_from_temperature(8000)
        # self.render_pipeline.add_light(light)

        light = PointLight()
        light.pos = -11.2, -13.84, -9.24
        light.radius = 1e20
        light.set_color_from_temperature(8000)
        light.energy = 2500
        # self.render_pipeline.add_light(light)

        # envprobe = self.render_pipeline.add_environment_probe()
        # envprobe.set_pos(0, -16.2, 4.4)
        # envprobe.set_scale(40, 40, 40)
        # envprobe.parallax_correction = False


Application()



import copy

from metadrive.component.traffic_light.nuplan_traffic_light import NuplanTrafficLight
from metadrive.manager.base_manager import BaseManager
from metadrive.scenario.scenario_description import ScenarioDescription as SD


class NuPlanLightManager(BaseManager):
    CLEAR_LIGHTS = False

    def __init__(self):
        super(NuPlanLightManager, self).__init__()
        self._lane_to_lights = {}
        self.nuplan_id_to_obj_id = {}
        self.obj_id_to_nuplan_id = {}

        self._episode_light_data = None

    def before_reset(self):
        super(NuPlanLightManager, self).before_reset()
        self._lane_to_lights = {}
        self.nuplan_id_to_obj_id = {}
        self.obj_id_to_nuplan_id = {}

        self._episode_light_data = self._get_episode_light_data()

    def after_reset(self):
        for light in self._episode_light_data[0]:
            lane_info = self.engine.current_map.road_network.graph[str(light.lane_connector_id)]
            traffic_light = self.spawn_object(NuplanTrafficLight, lane=lane_info.lane)
            self._lane_to_lights[lane_info.lane.index] = traffic_light
            self.nuplan_id_to_obj_id[str(light.lane_connector_id)] = traffic_light.name
            self.obj_id_to_nuplan_id[traffic_light.name] = str(light.lane_connector_id)
            traffic_light.set_status(light.status)

    def after_step(self, *args, **kwargs):
        if self.episode_step >= self.current_scenario_length:
            return

        step_data = self._episode_light_data[self.engine.episode_step]

        if self.CLEAR_LIGHTS:
            light_to_eliminate = self._lane_to_lights.keys() - set([str(i.lane_connector_id) for i in step_data])
            for lane_id in light_to_eliminate:
                light = self._lane_to_lights.pop(lane_id)
                nuplan_id = self.obj_id_to_nuplan_id.pop(light.id)
                obj_id = self.nuplan_id_to_obj_id.pop(nuplan_id)
                assert obj_id == light.name
                self.clear_objects([obj_id])

        for light in self._episode_light_data[self.episode_step]:
            if str(light.lane_connector_id) in self._lane_to_lights:
                traffic_light = self._lane_to_lights[str(light.lane_connector_id)]
            else:
                try:
                    lane_info = self.engine.current_map.road_network.graph[str(light.lane_connector_id)]
                except KeyError:
                    continue
                traffic_light = self.spawn_object(NuplanTrafficLight, lane=lane_info.lane)
                assert str(light.lane_connector_id) == lane_info.lane.index
                self._lane_to_lights[lane_info.lane.index] = traffic_light
                self.nuplan_id_to_obj_id[str(light.lane_connector_id)] = traffic_light.name
                self.obj_id_to_nuplan_id[traffic_light.name] = str(light.lane_connector_id)
            traffic_light.set_status(light.status)

        assert len(self._lane_to_lights) == len(self.nuplan_id_to_obj_id) == len(self.obj_id_to_nuplan_id)

    def has_traffic_light(self, lane_index):
        return True if lane_index in self._lane_to_lights else False

    @property
    def current_scenario(self):
        return self.engine.data_manager.current_scenario

    def traffic_light_status_at(self, timestep):
        return self.current_scenario.get_traffic_light_status_at_iteration(timestep)

    @property
    def current_scenario_length(self):
        return self.engine.data_manager.current_scenario_length

    def _get_episode_light_data(self):
        length = self.engine.data_manager.current_scenario.get_number_of_iterations()
        ret = {i: [t for t in self.traffic_light_status_at(i)] for i in range(length)}
        return ret

    def get_state(self):
        return {
            SD.OBJ_ID_TO_ORIGINAL_ID: copy.deepcopy(self.obj_id_to_nuplan_id),
            SD.ORIGINAL_ID_TO_OBJ_ID: copy.deepcopy(self.nuplan_id_to_obj_id)
        }



"""

RenderPipeline

Copyright (c) 2014-2016 tobspr <tobias.springer1@gmail.com>

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.

"""

import importlib
import collections

from metadrive.render_pipeline.rplibs.six import iteritems, itervalues
from metadrive.render_pipeline.rplibs.yaml import load_yaml_file

from direct.stdpy.file import listdir, isdir, join, open

from metadrive.render_pipeline.rpcore.rpobject import RPObject
from metadrive.render_pipeline.rpcore.native import NATIVE_CXX_LOADED
from metadrive.render_pipeline.rpcore.pluginbase.setting_types import make_setting_from_data
from metadrive.render_pipeline.rpcore.pluginbase.day_setting_types import make_daysetting_from_data


class PluginManager(RPObject):
    """ This class manages all plugins. It provides functionality to load plugin
    settings, trigger callbacks on plugins, initialize the plugin instances
    and much more. """
    def __init__(self, pipeline):
        """ Constructs a new manager with no plugins loaded. To load settings
        and plugins, call load(). """
        RPObject.__init__(self)
        self._pipeline = pipeline
        self.settings = {}
        self.day_settings = {}
        self.instances = {}
        self.enabled_plugins = set()

        # Used by the plugin configurator and to only load the required data
        self.requires_daytime_settings = True

    def load(self):
        """ Loads all plugins and their settings, and also constructs instances
        of the main plugin classes for all enabled plugins """
        self.debug("Loading plugin settings")
        self.load_base_settings("/$$rp/rpplugins")
        self.load_setting_overrides("/$$rpconfig/plugins.yaml")

        if self.requires_daytime_settings:
            self.load_daytime_overrides("/$$rpconfig/daytime.yaml")

        self.debug("Creating plugin instances ..")
        for plugin_id in self.settings:
            handle = self._load_plugin(plugin_id)
            if handle:
                self.instances[plugin_id] = handle
            else:
                self.disable_plugin(plugin_id)

    def disable_plugin(self, plugin_id):
        """ Disables a plugin, given its plugin_id. This will remove it from
        the list of enabled plugins, if it ever was there """
        self.warn("Disabling", plugin_id)
        if plugin_id in self.enabled_plugins:
            self.enabled_plugins.remove(plugin_id)
        for instance in list(self.instances.values()):
            if plugin_id in instance.required_plugins:
                self.disable_plugin(instance.plugin_id)
        # if plugin_id in self.instances:
        #     del self.instances[plugin_id]

    def unload(self):
        """ Unloads all plugins """
        self.debug("Unloading all plugins")
        self.instances = {}
        self.settings = {}
        self.day_settings = {}
        self.enabled_plugins = set()

    def update(self):
        """ Main update method """
        pass

    def load_base_settings(self, plugin_dir):
        """ Loads the base settings of all plugins, even of disabled plugins.
        This is required to verify all overrides. """
        for entry in listdir(plugin_dir):
            abspath = join(plugin_dir, entry)
            if isdir(abspath) and entry not in ("__pycache__", "plugin_prefab"):
                self.load_plugin_settings(entry, abspath)

    def load_plugin_settings(self, plugin_id, plugin_pth):
        """ Internal method to load all settings of a plugin, given its plugin
        id and path to the plugin base directory """
        config_file = join(plugin_pth, "config.yaml")
        config = load_yaml_file(config_file)
        # When you don't specify anything in the settings, instead of
        # returning an empty dictionary, pyyaml returns None
        config["settings"] = config["settings"] or []
        config["daytime_settings"] = config["daytime_settings"] or []

        if config["settings"] and len(config["settings"][0]) != 2:
            self.fatal("Invalid plugin configuration, did you miss '!!omap'?")

        settings = collections.OrderedDict([(k, make_setting_from_data(v)) for k, v in config["settings"]])
        self.settings[plugin_id] = settings

        if self.requires_daytime_settings:
            daysettings = collections.OrderedDict(
                [(k, make_daysetting_from_data(v)) for k, v in config["daytime_settings"]]
            )
            self.day_settings[plugin_id] = daysettings

    def load_setting_overrides(self, override_path):
        """ Loads an override file for the settings, which contains values to
        override the settings with """
        overrides = load_yaml_file(override_path)
        if not overrides:
            self.warn("Failed to load overrides")
            return
        self.enabled_plugins = set(overrides["enabled"] or [])
        for plugin_id, pluginsettings in iteritems(overrides["overrides"] or {}):
            if plugin_id not in self.settings:
                self.warn("Unkown plugin in plugin config:", plugin_id)
                continue
            for setting_id, setting_val in iteritems(pluginsettings or {}):
                if setting_id not in self.settings[plugin_id]:
                    self.warn("Unkown override:", plugin_id, ":", setting_id)
                    continue
                self.settings[plugin_id][setting_id].set_value(setting_val)

    def load_daytime_overrides(self, override_path):
        """ Loads an override file for the daytime settings, which contains
        values to override the settings with """
        overrides = load_yaml_file(override_path)
        if not overrides:
            self.warn("Failed to load daytime overrides")
            return
        for plugin_id, settings in iteritems(overrides["control_points"] or {}):
            for setting_id, control_points in iteritems(settings):
                if setting_id not in self.day_settings[plugin_id]:
                    self.warn("Unkown daytime override:", plugin_id, ":", setting_id)
                    continue
                self.day_settings[plugin_id][setting_id].set_control_points(control_points)

    def trigger_hook(self, hook_name):
        """ Triggers a given hook on all plugins, effectively calling all
        bound callbacks """
        hook_method = "on_{}".format(hook_name)
        for plugin_id in self.enabled_plugins:
            plugin_handle = self.instances[plugin_id]
            if hasattr(plugin_handle, hook_method):
                getattr(plugin_handle, hook_method)()

    def is_plugin_enabled(self, plugin_id):
        """ Returns whether a plugin is currently enabled and loaded """
        return plugin_id in self.enabled_plugins

    def get_setting_handle(self, plugin_id, setting_id):
        """ Returns the handle to a setting """
        return self.settings[plugin_id][setting_id]

    def init_defines(self):
        """ Initializes all plugin settings as a define, so they can be queried
        in a shader """
        for plugin_id in self.enabled_plugins:
            pluginsettings = self.settings[plugin_id]
            self._pipeline.stage_mgr.defines["HAVE_PLUGIN_{}".format(plugin_id)] = 1
            for setting_id, setting in iteritems(pluginsettings):
                if setting.shader_runtime or not setting.runtime:
                    # Only store settings which either never change, or trigger
                    # a shader reload when they change
                    setting.add_defines(plugin_id, setting_id, self._pipeline.stage_mgr.defines)

    def _load_plugin(self, plugin_id):
        """ Internal method to load a plugin """
        plugin_class = "metadrive.render_pipeline.rpplugins.{}.plugin".format(plugin_id)
        module = importlib.import_module(plugin_class)
        instance = module.Plugin(self._pipeline)
        if instance.native_only and not NATIVE_CXX_LOADED:
            if plugin_id in self.enabled_plugins:
                self.warn("Cannot load", plugin_id, "since it requires the C++ modules.")
                return False
        for required_plugin in instance.required_plugins:
            if required_plugin not in self.enabled_plugins:
                if plugin_id in self.enabled_plugins:
                    self.warn("Cannot load {} since it requires {}".format(plugin_id, required_plugin))
                    return False
                break
        return instance

    def save_overrides(self, override_file):
        """ Saves all overrides to the given file """
        output = "\n# Render Pipeline Plugin Configuration\n"
        output += "# Instead of editing this file, prefer to use the Plugin Configurator\n"
        output += "# Any formatting and comments will be lost\n\n"
        output += "enabled:\n"

        def sort_criteria(pid):
            return ("A" if self.is_plugin_enabled(pid) else "B") + pid

        for plugin_id in sorted(self.settings, key=sort_criteria):
            output += "   {}- {}\n".format(" # " if plugin_id not in self.enabled_plugins else " ", plugin_id)
        output += "\n\noverrides:\n"
        for plugin_id, pluginsettings in sorted(iteritems(self.settings)):
            output += " " * 4 + plugin_id + ":\n"
            for setting_id, setting_handle in iteritems(pluginsettings):
                output += " " * 8 + "{}: {}\n".format(setting_id, setting_handle.value)
            output += "\n"
        with open(override_file, "w") as handle:
            handle.write(output)

    def save_daytime_overrides(self, override_file):
        """ Saves all time of day overrides to the given file """
        output = "\n# Render Pipeline Time Of Day Configuration\n"
        output += "# Instead of editing this file, prefer to use the Time Of Day Editor\n"
        output += "# Any formatting and comments will be lost\n\n"
        output += "control_points:\n"
        for plugin_id, settings in iteritems(self.day_settings):
            if settings:
                output += " " * 4 + plugin_id + ":\n"
                for setting_id, setting_handle in iteritems(settings):
                    output += " " * 8 + setting_id + ": "
                    output += setting_handle.serialize() + "\n"
        with open(override_file, "w") as handle:
            handle.write(output)

    def set_plugin_enabled(self, plugin_id, enabled=True):
        """ Sets whether a plugin is enabled or not, thus should be loaded when
        the pipeline starts or not """
        if enabled:
            self.enabled_plugins.add(plugin_id)
        else:
            self.enabled_plugins.remove(plugin_id)

    def reset_plugin_settings(self, plugin_id):
        """ Resets all settings of a given plugin """
        for setting in itervalues(self.settings[plugin_id]):
            setting.value = setting.default

    def on_setting_changed(self, plugin_id, setting_id, value):
        """ Callback when a setting got changed. This will update the setting,
        and also call the callback for that setting, in case the plugin defined
        one. """
        if plugin_id not in self.settings or setting_id not in self.settings[plugin_id]:
            self.warn("Got invalid setting change:", plugin_id, "/", setting_id)
            return

        setting = self.settings[plugin_id][setting_id]
        setting.set_value(value)

        if plugin_id not in self.enabled_plugins:
            return

        if setting.runtime or setting.shader_runtime:
            update_method = self.instances[plugin_id], "update_" + setting_id
            if hasattr(*update_method):
                getattr(*update_method)()

        if setting.shader_runtime:
            self.init_defines()
            self._pipeline.stage_mgr.write_autoconfig()
            self.instances[plugin_id].reload_shaders()



from metadrive.utils.random import get_np_random


class Randomizable:
    """
    Global Random Engine serving as the super class of BaseManager, it can sync all seeds in all managers automatically
    Then all objects generated by XxxManager will be set a random seed by GlobalRandomEngine.randint().
    In the lifetime of GlobalRandomEngine's subclasses, these instances will be seeded for several times
    """
    MAX_RAND_INT = 65536

    def __init__(self, random_seed):
        self.random_seed = random_seed
        self.np_random = get_np_random(self.random_seed)

    def seed(self, random_seed):
        self.random_seed = random_seed
        self.np_random = get_np_random(random_seed)

    def generate_seed(self):
        return self.np_random.randint(0, self.MAX_RAND_INT)



import logging

logger = logging.getLogger(__name__)

logger.warning(
    "This converters will de deprecated. "
    "Use tools in ScenarioNet instead: https://github.com/metadriverse/ScenarioNet."
)



import open3d as o3d
import matplotlib.pyplot as plt
import numpy as np

if __name__ == '__main__':
    # 8bit 1channel or 8bit 3channel, and depth image is: 16bit 1channel image
    pcds = []
    # for h in range(-180, 180, 20):
    h = 0
    raw_depth = o3d.io.read_image("C:\\Users\\x1\\Desktop\\neurips_2023\\sensor_video\\depth\\depth_36.jpg")
    raw_depth = np.array(raw_depth)[..., 0]
    raw_depth *= 255
    raw_depth = raw_depth.astype(np.uint16)
    raw_depth = o3d.geometry.Image(raw_depth)
    extrinsic = np.array([[np.cos(h), -np.sin(h), 0, 0], [np.sin(h), np.cos(h), 0, 0], [0, 0, 1, 0], [0, 0, 0, 1]])
    # extrinsic = np.array([[1, 0, 0, 0],
    #                       [0, np.cos(h), -np.sin(h), 0],
    #                       [0, np.sin(h), np.cos(h), 1.5],
    #                       [0, 0, 0, 1]])
    intrinsic = o3d.camera.PinholeCameraIntrinsic(o3d.camera.PinholeCameraIntrinsicParameters.PrimeSenseDefault)
    intrinsic.set_intrinsics(1600, 900, 0.866, 0.866, 800, 450)
    pcd = o3d.geometry.PointCloud.create_from_depth_image(
        depth=raw_depth,
        intrinsic=intrinsic,
        extrinsic=extrinsic,
        project_valid_depth_only=False,
        stride=4
        # depth_scale=1000.0,
        # depth_trunc=1000.0
    )
    # pcd.transform([[1, 0, 0, 0], [0, -1, 0, 0], [0, 0, -1, 0], [0, 0, 0, 1]])
    r = pcd.get_rotation_matrix_from_xyz((0, 0, np.pi))
    pcd = pcd.rotate(r, center=(0, 0, 0))
    pcds.append(pcd)

    o3d.visualization.draw_geometries(pcds)

    # CX_DEPTH=0
    # CY_DEPTH=0
    # FX_DEPTH=300
    # FY_DEPTH=300
    # raw_depth = o3d.io.read_image('debug_{}.jpg'.format(0))
    # raw_depth = np.array(raw_depth)[..., 0]
    # raw_depth = raw_depth.astype(np.uint8)
    # width = 800
    # height = 600
    # # o3d.visualization.draw_geometries(pcds)
    #
    # jj = np.tile(range(width), height)
    # ii = np.repeat(range(height), width)
    # # Compute constants:
    # xx = (jj - CX_DEPTH) / FX_DEPTH
    # yy = (ii - CY_DEPTH) / FY_DEPTH
    # # transform depth image to vector of z:
    # length = height * width
    # z = raw_depth.reshape(height * width)
    # # compute point cloud
    # pcd = np.dstack((xx * z, yy * z, z)).reshape((length, 3))
    # pcd_o3d = o3d.geometry.PointCloud()  # create point cloud object
    # pcd_o3d.points = o3d.utility.Vector3dVector(pcd)  # set pcd_np as the point cloud points
    # # Visualize:
    # o3d.visualization.draw_geometries([pcd_o3d])



"""

RenderPipeline

Copyright (c) 2014-2016 tobspr <tobias.springer1@gmail.com>

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.

"""

from direct.stdpy.file import open

from metadrive.render_pipeline.rpcore.render_stage import RenderStage
from metadrive.render_pipeline.rpcore.globals import Globals


class UpdatePreviousPipesStage(RenderStage):
    """ This stage is constructed by the StageManager and stores all the
    current pipes in the previous pipe storage.

    This stage is a bit special, and not like the other stages, it does not
    specify inputs, since the StageManager passes all required inputs on demand.
    Also this stage does not load any shaders, but creates them on the fly.
    """
    def __init__(self, pipeline):
        RenderStage.__init__(self, pipeline)
        self._transfers = []

    def add_transfer(self, from_tex, to_tex):
        """ Adds a new texture which should be copied from "from_tex" to
        "to_tex". This should be called before the stage gets constructed.
        The source texture is expected to have the same size as the render
        resolution. """
        self._transfers.append((from_tex, to_tex))

    def create(self):
        self.debug("Creating previous pipes stage ..")
        self._target = self.create_target("StorePreviousPipes")
        self._target.prepare_buffer()

        # Set inputs
        for i, (from_tex, to_tex) in enumerate(self._transfers):  # pylint: disable=unused-variable
            self._target.set_shader_input("SrcTex" + str(i), from_tex)
            self._target.set_shader_input("DestTex" + str(i), to_tex)

    def set_dimensions(self):
        """ Sets the dimensions on all targets. See RenderTarget::set_dimensions """
        for from_tex, to_tex in self._transfers:  # pylint: disable=unused-variable
            to_tex.set_x_size(Globals.resolution.x)
            to_tex.set_y_size(Globals.resolution.y)

    def reload_shaders(self):
        """ This method augo-generates a shader which copies all textures specified
        as "from-tex" to the textures specified as "to-tex". """
        uniforms = []
        lines = []

        # Collect all samplers and generate the required uniforms and copy code
        for i, (from_tex, to_tex) in enumerate(self._transfers):
            index = str(i)
            uniforms.append(self.get_sampler_type(from_tex) + " SrcTex" + index)
            uniforms.append(self.get_sampler_type(to_tex, True) + " DestTex" + index)
            lines += [
                "\n  // Copying " + from_tex.get_name() + " to " + to_tex.get_name(),
                self.get_sampler_lookup(from_tex, "data" + index, "SrcTex" + index, "coord_2d_int"),
                self.get_store_code(to_tex, "DestTex" + index, "coord_2d_int", "data" + index),
                "\n",
            ]

        # Actually create the shader
        fragment = "#version 430\n"
        fragment += "\n// Autogenerated, do not edit! Your changes will be lost.\n\n"
        for uniform in uniforms:
            fragment += "uniform " + uniform + ";\n"
        fragment += "\nvoid main() {\n"
        fragment += "  ivec2 coord_2d_int = ivec2(gl_FragCoord.xy);\n"
        for line in lines:
            fragment += "  " + line + "\n"
        fragment += "}\n"

        # Write the shader
        shader_dest = "/$$rptemp/$$update_previous_pipes.frag.glsl"
        with open(shader_dest, "w") as handle:
            handle.write(fragment)

        # Load it back again
        self._target.shader = self.load_shader(shader_dest)

    def get_sampler_type(self, tex, can_write=False):  # pylint: disable=unused-argument
        """ Returns the matching GLSL sampler type for a Texture, or image type
        in case write access is required """
        # TODO: Add more sampler types based on texture type
        if not can_write:
            return "sampler2D"
        else:
            return "writeonly image2D"

    def get_sampler_lookup(self, tex, dest_name, sampler_name, coord_var):  # noqa # pylint: disable=unused-argument
        """ Returns the matching GLSL sampler lookup for a texture, storing the
        result in the given glsl variable """
        # TODO: Add more lookups based on texture type
        return "vec4 " + dest_name + " = texelFetch(" + sampler_name + ", " + coord_var + ", 0);"

    def get_store_code(self, tex, sampler_name, coord_var, data_var):  # noqa# pylint: disable=unused-argument
        """ Returns the matching GLSL code to store the given data in a given
        texture """
        # TODO: Add more stores based on texture type
        return "imageStore(" + sampler_name + ", " + coord_var + ", vec4(" + data_var + "));"



import numpy as np

from metadrive.component.vehicle.vehicle_type import DefaultVehicle
from metadrive.engine.engine_utils import initialize_engine
from metadrive.envs import MetaDriveEnv
from metadrive.envs.base_env import BASE_DEFAULT_CONFIG
from metadrive.envs.metadrive_env import METADRIVE_DEFAULT_CONFIG
from metadrive.policy.idm_policy import IDMPolicy
from metadrive.utils import Config


def _create_vehicle():
    v_config = Config(BASE_DEFAULT_CONFIG["vehicle_config"]).update(METADRIVE_DEFAULT_CONFIG["vehicle_config"])
    v_config.update({"use_render": False, "image_observation": False})
    config = Config(BASE_DEFAULT_CONFIG)
    config.update(
        {
            "use_render": False,
            "pstats": False,
            "image_observation": False,
            "debug": False,
            "vehicle_config": v_config
        }
    )
    initialize_engine(config)
    v = DefaultVehicle(vehicle_config=v_config, random_seed=0)
    return v


def test_idm_policy_briefly():
    env = MetaDriveEnv()
    env.reset()
    try:
        vehicles = env.engine.traffic_manager.traffic_vehicles
        for v in vehicles:
            policy = IDMPolicy(
                vehicle=v, traffic_manager=env.engine.traffic_manager, delay_time=1, random_seed=env.current_seed
            )
            action = policy.before_step(v, front_vehicle=None, rear_vehicle=None, current_map=env.engine.current_map)
            action = policy.step(dt=0.02)
            action = policy.after_step(v, front_vehicle=None, rear_vehicle=None, current_map=env.engine.current_map)
            env.engine.policy_manager.register_new_policy(
                IDMPolicy,
                vehicle=v,
                traffic_manager=env.engine.traffic_manager,
                delay_time=1,
                random_seed=env.current_seed
            )
        env.step(env.action_space.sample())
        env.reset()
    finally:
        env.close()


def test_idm_policy_is_moving(render=False, in_test=True):
    # config = {"traffic_mode": "hybrid", "map": "SS", "traffic_density": 1.0}
    config = {"traffic_mode": "respawn", "map": "SS", "traffic_density": 1.0}
    if render:
        config.update({"use_render": True, "manual_control": True})
    env = MetaDriveEnv(config)
    env.reset(seed=0)
    last_pos = None
    try:
        for t in range(100):
            env.step(env.action_space.sample())
            vs = env.engine.traffic_manager.traffic_vehicles
            # # print("Position: ", {str(v)[:4]: v.position for v in vs})
            new_pos = np.array([v.position for v in vs])
            if t > 50 and last_pos is not None and in_test:
                assert np.any(new_pos != last_pos)
            last_pos = new_pos
        env.reset()
    finally:
        env.close()


if __name__ == '__main__':
    # test_idm_policy_briefly()
    test_idm_policy_is_moving(render=True, in_test=False)



# Copyright (c) 2020 kitsune.ONE team.

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import bpy


def get_object_collections(obj):
    results = []

    for collection in bpy.data.collections:
        if collection.name == 'RigidBodyWorld':
            continue

        for obj2 in collection.objects:
            if obj2.name == obj.name:
                results.append(collection)

    return results


def get_object_collection(obj):
    collections = get_object_collections(obj)
    if collections:
        return collections[0]



import copy
from metadrive.envs.marl_envs.multi_agent_metadrive import MULTI_AGENT_METADRIVE_DEFAULT_CONFIG
MULTI_AGENT_METADRIVE_DEFAULT_CONFIG["force_seed_spawn_manager"] = True
from metadrive.envs.marl_envs.marl_inout_roundabout import MultiAgentRoundaboutEnv


def test_ma_env_force_reset():
    def close_and_reset_num_agents(env, num_agents, raw_input_config):
        config = copy.deepcopy(raw_input_config)
        env.close()
        config["num_agents"] = num_agents
        env.__init__(config)

    config = {'num_agents': 1}
    e = MultiAgentRoundaboutEnv(config)
    _raw_input_config = copy.deepcopy(config)
    e.reset()
    assert len(e.vehicles) == e.num_agents == len(e.config["target_vehicle_configs"]) == 1

    close_and_reset_num_agents(e, num_agents=2, raw_input_config=_raw_input_config)
    e.reset()
    assert len(e.vehicles) == e.num_agents == len(e.config["target_vehicle_configs"]) == 2

    close_and_reset_num_agents(e, num_agents=5, raw_input_config=_raw_input_config)
    e.reset()
    assert len(e.vehicles) == e.num_agents == len(e.config["target_vehicle_configs"]) == 5

    e.close()


if __name__ == '__main__':
    test_ma_env_force_reset()
